from typing import Dict, Any
import edustudio.settings as settings
from edustudio.utils.common import IDUtil as idUtil
from edustudio.utils.common import UnifyConfig
import argparse
from ast import literal_eval
from collections import defaultdict
import importlib


def get_global_cfg(
    dataset:str,
    cfg_file_name:str,
    traintpl_cfg_dict: Dict[str, Any],
    datatpl_cfg_dict:  Dict[str, Any],
    evaltpl_cfg_dict:  Dict[str, Any],
    modeltpl_cfg_dict: Dict[str, Any],
    frame_cfg_dict:  Dict[str, Any],
):
    """merge configurations from different entrypoint into a global config object

    Args:
        dataset (str): dataset name
        cfg_file_name (str): config file name
        traintpl_cfg_dict (Dict[str, Any]): parameter dict of training template
        datatpl_cfg_dict (Dict[str, Any]): parameter dict of data template
        evaltpl_cfg_dict (Dict[str, Any]): parameter dict of evaluate template
        modeltpl_cfg_dict (Dict[str, Any]): parameter dict of model template
        frame_cfg_dict (Dict[str, Any]): parameter dict of framework template

    Returns:
        UnifyConfig: the global config object
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('--dataset', '-dt', type=str,
                        help='dataset name', dest='dataset', default=dataset)
    parser.add_argument('--cfg_file_name', '-f', type=str,
                        help='config filename', dest='cfg_file_name', default=cfg_file_name)
    parser.add_argument('--traintpl_cfg.cls', '-traintpl_cls', type=str,
                        dest='traintpl_cls', default=traintpl_cfg_dict.get('cls', None))
    parser.add_argument('--datatpl_cfg.cls', '-datatpl_cls', type=str,
                        dest='datatpl_cls', default=datatpl_cfg_dict.get('cls', None))
    parser.add_argument('--modeltpl_cfg.cls', '-modeltpl_cls', type=str,
                        dest='modeltpl_cls', default=modeltpl_cfg_dict.get('cls', None))
    parser.add_argument('--evaltpl_cfg.clses', '-evaltpl_clses', nargs='+',
                        dest='evaltpl_clses', default=evaltpl_cfg_dict.get('clses', None))
    parser.add_argument('--datatpl_cfg.backbone_datatpl_cls', '-datatpl_backbone_cls', type=str,
                        dest='backbone_datatpl_cls', default=datatpl_cfg_dict.get('backbone_datatpl_cls', None))
    parser.add_argument('--modeltpl_cfg.backbone_modeltpl_cls', '-modeltpl_backbone_cls', type=str,
                        dest='backbone_modeltpl_cls', default=modeltpl_cfg_dict.get('backbone_modeltpl_cls', None))
    parser.add_argument('--datatpl_cfg.mid2cache_op_seq', '-mid2cache_op_seq', type=str,
                        dest='mid2cache_op_seq', default=datatpl_cfg_dict.get('mid2cache_op_seq', None))   

    try:
        __IPYTHON__
        _default_args = []
    except NameError:
        _default_args = None
    args, unknown_args = parser.parse_known_args(args=_default_args)
    assert args.dataset is not None
    
    unknown_arg_dict = defaultdict(dict)
    if len(unknown_args) > 0:
        assert len(unknown_args) % 2 == 0, \
            "number of extra parameters[except dt and cfg_file_name] from command line should be even"
        for i in range(int(len(unknown_args) / 2)):
            assert unknown_args[2 * i].startswith("--"), \
                "the first term in extra parameter[except dt and cfg_file_name] pair should start with '--'"
            key, value = unknown_args[2 * i][2:], unknown_args[2 * i + 1]
            if key.startswith('traintpl_cfg.'):
                unknown_arg_dict['traintpl_cfg'][key] = value
            elif key.startswith('datatpl_cfg.'):
                unknown_arg_dict['datatpl_cfg'][key] = value
            elif key.startswith('modeltpl_cfg.'):
                unknown_arg_dict['modeltpl_cfg'][key] = value
            elif key.startswith('evaltpl_cfg.'):
               unknown_arg_dict['evaltpl_cfg'][key] = value
            elif key.startswith('frame_cfg.'):
                unknown_arg_dict['frame_cfg'][key] = value
            else:
                pass
                # raise ValueError(f"unsupported key: {key}")


    cfg = UnifyConfig({
        'traintpl_cfg': UnifyConfig(), 'datatpl_cfg': UnifyConfig(),
        'modeltpl_cfg': UnifyConfig(), 'evaltpl_cfg': UnifyConfig(), 
        'frame_cfg': UnifyConfig()
    })
    cfg.dataset = args.dataset
        
    # process frame cfg
    cfg.frame_cfg = UnifyConfig.from_py_module(settings)
    cfg.frame_cfg.ID = idUtil.get_random_id_bytime()
    for k,v in frame_cfg_dict.items():
        assert k in cfg.frame_cfg, f"invalid key: {k}"
        assert type(v) is None or type(cfg.frame_cfg[k]) is type(v)
        cfg.frame_cfg[k] = v
    for k,v in unknown_arg_dict['frame_cfg'].items():
        assert k in cfg.frame_cfg, f"invalid key: {k}"
        if type(cfg.frame_cfg[k]) is not str:
            v = type(cfg.frame_cfg)(literal_eval(v))
        cfg.dot_set(k, v)

    traintpl_cls = args.traintpl_cls
    datatpl_cls = args.datatpl_cls
    modeltpl_cls = args.modeltpl_cls
    evaltpl_clses = args.evaltpl_clses
    if args.cfg_file_name is not None:
        yaml_cfg = UnifyConfig.from_yml_file(
            f"{cfg.frame_cfg['CFG_FOLDER_PATH']}/{cfg.dataset}/{args.cfg_file_name}"
        )
        assert 'frame_cfg' not in yaml_cfg
        assert 'dataset' not in yaml_cfg
        assert 'logger' not in yaml_cfg
        traintpl_cls = traintpl_cls or yaml_cfg.get('traintpl_cfg', {'cls': None}).get("cls")
        datatpl_cls = datatpl_cls or yaml_cfg.get('datatpl_cfg', {'cls': None}).get("cls")
        modeltpl_cls = modeltpl_cls or yaml_cfg.get('modeltpl_cfg', {'cls': None}).get("cls")
        evaltpl_clses = evaltpl_clses or yaml_cfg.get('evaltpl_cfg', {'clses': None}).get("clses")

        args.backbone_modeltpl_cls = args.backbone_modeltpl_cls or yaml_cfg.get('modeltpl_cfg', {'backbone_modeltpl_cls': None}).get("backbone_modeltpl_cls")
        args.backbone_datatpl_cls = args.backbone_datatpl_cls or yaml_cfg.get('datatpl_cfg', {'backbone_datatpl_cls': None}).get("backbone_datatpl_cls")
        args.mid2cache_op_seq = args.mid2cache_op_seq or yaml_cfg.get('datatpl_cfg', {'mid2cache_op_seq': None}).get("mid2cache_op_seq")

    assert traintpl_cls is not None
    assert datatpl_cls is not None
    assert modeltpl_cls is not None
    assert evaltpl_clses is not None
    assert len(set(evaltpl_clses)) == len(evaltpl_clses)


    cfg.dot_set('traintpl_cfg.cls', traintpl_cls)
    cfg.dot_set('datatpl_cfg.cls', datatpl_cls)
    cfg.dot_set('modeltpl_cfg.cls', modeltpl_cls)
    cfg.dot_set('evaltpl_cfg.clses', evaltpl_clses)
    
    if isinstance(traintpl_cls, str):
        cfg.traintpl_cfg.update(
            importlib.import_module("edustudio.traintpl").
            __getattribute__(traintpl_cls).get_default_cfg()
        )
    else:
        cfg.traintpl_cfg.update(traintpl_cls.get_default_cfg())
    
    if isinstance(modeltpl_cls, str):
        cfg.modeltpl_cfg.update(
            importlib.import_module("edustudio.model").
            __getattribute__(modeltpl_cls).get_default_cfg(backbone_modeltpl_cls=args.backbone_modeltpl_cls)
        )
    else:
        cfg.modeltpl_cfg.update(modeltpl_cls.get_default_cfg(backbone_modeltpl_cls=args.backbone_modeltpl_cls))
    
    if isinstance(datatpl_cls, str):
        cfg.datatpl_cfg.update(
            importlib.import_module("edustudio.datatpl").
            __getattribute__(datatpl_cls).get_default_cfg(backbone_datatpl_cls=args.backbone_datatpl_cls, mid2cache_op_seq=args.mid2cache_op_seq)
        )
    else:
        cfg.datatpl_cfg.update(datatpl_cls.get_default_cfg(backbone_datatpl_cls=args.backbone_datatpl_cls, mid2cache_op_seq=args.mid2cache_op_seq))
    
    for evaltpl_cls in evaltpl_clses:
        if isinstance(evaltpl_cls, str):
            cfg.evaltpl_cfg[evaltpl_cls] = importlib.import_module("edustudio.evaltpl").\
                __getattribute__(evaltpl_cls).get_default_cfg()
        else:
            cfg.evaltpl_cfg[evaltpl_cls.__name__] = evaltpl_cls.get_default_cfg()


    if args.mid2cache_op_seq is not None:
        atom_data_op_set = {(op if type(op) is str else op.__name__) for op in args.mid2cache_op_seq}
    else:
        atom_data_op_set = {(op if type(op) is str else op.__name__) for op in cfg['datatpl_cfg'].get('mid2cache_op_seq', [])}

    # config file
    if args.cfg_file_name is not None:
        for config_name in ['traintpl_cfg', 'datatpl_cfg', 'modeltpl_cfg', 'evaltpl_cfg']:
            if config_name not in yaml_cfg: continue
            if config_name == 'datatpl_cfg':
                for k,v in yaml_cfg[config_name].items():
                    assert k in cfg[config_name], f"invalid key: {k}"
                    if k == 'cls': continue
                    # assert type(v) is None or type(cfg[config_name][k]) is type(v)
                    if k in atom_data_op_set:
                        for kk,vv in yaml_cfg[config_name].get(k, {}).items():
                            assert kk in cfg[config_name][k], f"invalid key: {kk}"
                            cfg[config_name][k][kk] = vv
                    else:
                        cfg[config_name][k] = v 
            if config_name in ['traintpl_cfg', 'modeltpl_cfg']:
                for k,v in yaml_cfg[config_name].items():
                    assert k in cfg[config_name], f"invalid key: {k}"
                    if k == 'cls': continue
                    # assert type(v) is None or type(cfg[config_name][k]) is type(v)
                    cfg[config_name][k] = v
            if config_name in ['evaltpl_cfg']:
                for k,v in yaml_cfg[config_name].items():
                    if k == 'clses': continue
                    assert k in cfg.evaltpl_cfg['clses'], f"invalid key: {k}"
                    assert type(v) is dict
                    for kk, vv in v.items():
                        assert kk in cfg.evaltpl_cfg[k], f"invalid key: {kk}"
                        assert type(cfg.evaltpl_cfg[k][kk]) is type(vv)
                        cfg.evaltpl_cfg[k][kk] = vv

    # parameter dict
    for k,v in traintpl_cfg_dict.items():
        if k == 'cls': continue
        assert k in cfg['traintpl_cfg'], f"invalid key: {k}"
        # assert type(v) is None or type(cfg['traintpl_cfg'][k]) is type(v)
        cfg['traintpl_cfg'][k] = v
    
    for k,v in datatpl_cfg_dict.items():
        if k == 'cls' or k == 'backbone_datatpl_cls': continue
        assert k in cfg['datatpl_cfg'], f"invalid key: {k}"
        # assert type(v) is None or type(cfg['datatpl_cfg'][k]) is type(v)
        if k in atom_data_op_set:
            for kk,vv in datatpl_cfg_dict[k].items():
                assert kk in cfg['datatpl_cfg'][k], f"invalid key: {kk}"
                cfg['datatpl_cfg'][k][kk] = vv
        else:
            cfg['datatpl_cfg'][k] = v
    for k,v in modeltpl_cfg_dict.items():
        if k == 'cls' or k == 'backbone_modeltpl_cls': continue
        assert k in cfg['modeltpl_cfg'], f"invalid key: {k}"
        # assert type(v) is None or type(cfg['modeltpl_cfg'][k]) is type(v)
        cfg['modeltpl_cfg'][k] = v

    evaltpl_clses_name = [cls_ if isinstance(cls_, str) else cls_.__name__ for cls_ in cfg.evaltpl_cfg['clses']]
    for k,v in evaltpl_cfg_dict.items():
        if k == 'clses': continue
        assert k in evaltpl_clses_name, f"invalid key: {k}"
        assert type(v) is dict
        for kk, vv in v.items():
            assert kk in cfg.evaltpl_cfg[k], f"invalid key: {kk}"
            # assert type(cfg.evaltpl_cfg[k][kk]) is type(vv)
            cfg.evaltpl_cfg[k][kk] = vv
    

    # command line
    for config_name in ['traintpl_cfg', 'datatpl_cfg', 'modeltpl_cfg']:
        for k,v in unknown_arg_dict[config_name].items():
            if k == 'cls': continue
            kk = k.split('.')[1]
            assert kk in cfg[config_name], f"invalid key: {kk}"
            # if type(cfg[config_name][kk]) is not str:
            #     v = type(cfg[config_name][kk])(literal_eval(v))
            if type(cfg.dot_get(k)) is not str:
                v = type(cfg.dot_get(k))(literal_eval(v))
            cfg.dot_set(k, v)
    for k,v in unknown_arg_dict["evaltpl_cfg"].items():
        kk = k.split('.')[-1]
        assert kk in cfg['evaltpl_cfg'], f"invalid key: {kk}"
        if kk == 'clses': continue
        if type(cfg.dot_get(k)) is not str:
            v = type(cfg.dot_get(k))(literal_eval(v))
        cfg.dot_set(k, v)
    
    return cfg



r"""
NCDM
##########################################

Reference:
    Fei Wang et al. "Neural Cognitive Diagnosis for Intelligent Education Systems" in AAAI 2020.

Reference Code:
    https://github.com/bigdata-ustc/Neural_Cognitive_Diagnosis-NeuralCD
    https://github.com/bigdata-ustc/EduCDM/tree/main/EduCDM/NCDM

"""

from typing import Any
from ..gd_basemodel import GDBaseModel
import torch.nn as nn
from ..utils.components import PosMLP
import torch
import torch.nn.functional as F


class NCDM(GDBaseModel):
    r"""
    NCDM

    default_cfg:
        'dnn_units': [512, 256]  # dimension list of hidden layer in prediction layer
        'dropout_rate': 0.5      # dropout rate
        'activation': 'sigmoid'  # activation function in prediction layer
        'disc_scale': 10         # discrimination scale
    """
    
    default_cfg = {
        'dnn_units': [512, 256],
        'dropout_rate': 0.5,
        'activation': 'sigmoid',
        'disc_scale': 10
    }

    def __init__(self, cfg):
        super().__init__(cfg)

    def add_extra_data(self, **kwargs):
        self.Q_mat = kwargs['Q_mat'].to(self.device)

    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        self.n_cpt = self.datatpl_cfg['dt_info']['cpt_count']

    def build_model(self):
        # prediction sub-net
        self.student_emb = nn.Embedding(self.n_user, self.n_cpt)
        self.k_difficulty = nn.Embedding(self.n_item, self.n_cpt)
        self.e_difficulty = nn.Embedding(self.n_item, 1)
        self.pd_net = PosMLP(
            input_dim=self.n_cpt, output_dim=1, activation=self.modeltpl_cfg['activation'],
            dnn_units=self.modeltpl_cfg['dnn_units'], dropout_rate=self.modeltpl_cfg['dropout_rate']
        )

    def forward(self, stu_id, exer_id, **kwargs):
        # before prednet
        stu_emb = self.student_emb(stu_id)
        stat_emb = torch.sigmoid(stu_emb)
        k_difficulty = torch.sigmoid(self.k_difficulty(exer_id))
        e_difficulty = torch.sigmoid(self.e_difficulty(exer_id)) * self.modeltpl_cfg['disc_scale']
        # prednet
        input_knowledge_point = self.Q_mat[exer_id]
        input_x = e_difficulty * (stat_emb - k_difficulty) * input_knowledge_point
        pd = self.pd_net(input_x).sigmoid()
        return pd

    @torch.no_grad()
    def predict(self, stu_id, exer_id, **kwargs):
        return {
            'y_pd': self(stu_id, exer_id).flatten(),
        }

    def get_main_loss(self, **kwargs):
        stu_id = kwargs['stu_id']
        exer_id = kwargs['exer_id']
        label = kwargs['label']
        pd = self(stu_id, exer_id).flatten()
        loss = F.binary_cross_entropy(input=pd, target=label)
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)
    
    def get_stu_status(self, stu_id=None):
        if stu_id is not None:
            return self.student_emb(stu_id)
        else:
            return self.student_emb.weight



from typing import Dict, Any
import importlib
import sys
import logging
import shutil
import traceback

from .parse_cfg import get_global_cfg
from .init_all import init_all

from edustudio.utils.common import UnifyConfig


def run_edustudio(
    dataset: str = None,
    cfg_file_name: str = None,  # config file
    traintpl_cfg_dict: Dict[str, Any] = {},  # parameters dictionary
    datatpl_cfg_dict:  Dict[str, Any] = {},  # parameters dictionary
    evaltpl_cfg_dict:  Dict[str, Any] = {},  # parameters dictionary
    modeltpl_cfg_dict: Dict[str, Any] = {},  # parameters dictionary
    frame_cfg_dict:  Dict[str, Any] = {},  # parameters dictionary
    return_cfg_and_result: bool = False,
):
    """The quick start API to run edustudio

    Args:
        dataset (str, optional):dataset name. Defaults to None.
        cfg_file_name (str, optional): config file name. Defaults to None.
        traintpl_cfg_dict (Dict[str, Any]): parameter dict of training template
        datatpl_cfg_dict (Dict[str, Any]): parameter dict of data template
        evaltpl_cfg_dict (Dict[str, Any]): parameter dict of evaluate template
        modeltpl_cfg_dict (Dict[str, Any]): parameter dict of model template
        frame_cfg_dict (Dict[str, Any]): parameter dict of framework template

    Returns:
        tuple: the global config object and experimental result
    """
    cfg: UnifyConfig = get_global_cfg(
        dataset, cfg_file_name, traintpl_cfg_dict,
        datatpl_cfg_dict, evaltpl_cfg_dict, modeltpl_cfg_dict, frame_cfg_dict
    )
    init_all(cfg)
    ret = None
    try:
        cfg.logger.info("====" * 15)
        cfg.logger.info(f"[ID]: {cfg.frame_cfg.ID}")
        cfg.logger.info(f"[DATASET]: {cfg.dataset}")
        cfg.logger.info(f"[ARGV]: {sys.argv}")
        cfg.logger.info(f"[ALL_CFG]: \n{cfg.dump_fmt()}")
        cfg.dump_file(f"{cfg.frame_cfg.temp_folder_path}/cfg.json")
        cfg.logger.info("====" * 15)
        if isinstance(cfg.traintpl_cfg['cls'], str):
            cls = importlib.import_module('edustudio.traintpl').\
                __getattribute__(cfg.traintpl_cfg['cls'])
        else:
            cls = cfg.traintpl_cfg['cls']
        traintpl = cls(cfg)
        traintpl.start()
        cfg.logger.info(f"Task: {cfg.frame_cfg.ID} Completed!")
        
        for handler in cfg.logger.handlers:
            handler.close()
            cfg.logger.removeHandler(handler)

        for handler in cfg.logger.handlers:
            handler.close()
            cfg.logger.removeHandler(handler)

        shutil.move(cfg.frame_cfg.temp_folder_path, cfg.frame_cfg.archive_folder_path)
    except Exception as e:
        cfg.logger.error(traceback.format_exc())
        raise e
    
    if return_cfg_and_result: ret = (cfg, read_exp_result(cfg))
    return ret

def read_exp_result(cfg):
    with open(f"{cfg.frame_cfg.archive_folder_path}/{cfg.frame_cfg.ID}/result.json", 'r', encoding='utf-8') as f:
        import json
        return json.load(f)


if __name__ == "__main__":
    run_edustudio()



r"""
EKT
##########################################

Reference:
    Qi Liu et al. "EKT: Exercise-Aware Knowledge Tracing for Student Performance Prediction" in TKDE 2019.

Reference Code:
    https://github.com/bigdata-ustc/ekt

"""

from .eernn import EERNNM, EERNNA
import torch.nn as nn
import torch
import torch.nn.functional as F


class KnowledgeModel(nn.Module):
    """
    Transform Knowledge index to knowledge embedding
    """

    def __init__(self, know_len, know_emb_size):
        super(KnowledgeModel, self).__init__()
        self.knowledge_embedding = nn.Embedding(know_len+1, know_emb_size, padding_idx=0)

    def forward(self, cpt_seq, cpt_seq_mask):
        t = cpt_seq + 1
        t[cpt_seq_mask == 0] = 0
        return self.knowledge_embedding(t).sum(dim=1)


class EKTM(EERNNM):
    default_cfg = {
        'd_k': 25,
    }

    def build_cfg(self):
        super().build_cfg()
        self.n_cpt = self.datatpl_cfg['dt_info']['cpt_count']
        self.d_k = self.modeltpl_cfg['d_k']

    def build_model(self):
        super().build_model()
        self.cpt_mem_emb = nn.Embedding(self.n_cpt, self.d_k)
        self.W_K = nn.Embedding(self.n_cpt, self.d_k)
        self.h_initial = nn.Parameter(torch.zeros(self.n_cpt, self.d_h))
        self.knowledge_model = KnowledgeModel(self.n_cpt, self.d_k)
        self.stu_inter_seq_model = nn.GRU(4 * self.d_v, self.d_h)

    def _init_params(self):
        super()._init_params()
        self.h_initial.data.uniform_(-1, 1)

    def forward(self, exer_seq, label_seq, cpt_seq, cpt_seq_mask, **kwargs):
        exers, idx = exer_seq.unique(return_inverse=True)
        exer_emb = self.get_exer_embedding(exers)[idx]
        fused_exer_emb = exer_emb.repeat(1,1,2)
        cond = torch.where(label_seq == 1)
        fused_exer_emb[cond[0], cond[1], self.d_v*2:] = 0.0
        cond = torch.where(label_seq == 0)
        fused_exer_emb[cond[0], cond[1], 0:self.d_v*2] = 0.0


        seq_len = exer_seq.shape[1]
        h = self.h_initial
        pd_list = []
        for t in range(seq_len-1): # iter each time
            v_t = self.knowledge_model(cpt_seq[:,t,:], cpt_seq_mask[:,t,:])
            beta = torch.mm(self.cpt_mem_emb.weight, v_t.T).T.softmax(dim=1)

            # 计算预测值
            s = torch.mm(beta, h) # B*K, K*d -> B*d
            pred_v = torch.cat([exer_emb[:, t+1, :], s], dim=-1) # 1*B, 即每个学生在此刻对下一题的预测
            pd_t = self.pd_layer(pred_v)

            # 得到hidden vector
            xk = beta.unsqueeze(dim=-1) * fused_exer_emb[:,[t],:]
            _, h = self.stu_inter_seq_model(xk, h.unsqueeze(0))
            h = h.squeeze(0)
            pd_list.append(pd_t)
        return torch.concat(pd_list, dim=-1)


class EKTA(EKTM):
    def forward(self, exer_seq, label_seq, cpt_seq, cpt_seq_mask, **kwargs):
        exers, idx = exer_seq.unique(return_inverse=True)
        exer_emb = self.get_exer_embedding(exers)[idx]
        fused_exer_emb = exer_emb.repeat(1,1,2)
        cond = torch.where(label_seq == 1)
        fused_exer_emb[cond[0], cond[1], self.d_v*2:] = 0.0
        cond = torch.where(label_seq == 0)
        fused_exer_emb[cond[0], cond[1], 0:self.d_v*2] = 0.0

        seq_len = exer_seq.shape[1]
        h = self.h_initial
        pd_list = []
        h_list = [h]
        for t in range(seq_len-1): # iter each time
            v_t = self.knowledge_model(cpt_seq[:,t,:], cpt_seq_mask[:,t,:])
            beta = torch.mm(self.cpt_mem_emb.weight, v_t.T).T.softmax(dim=1)

            # 计算预测值
            cos_sim = F.cosine_similarity(exer_emb[:, [t+1], :], exer_emb[:, 0:t+1, :], dim=-1)
            h_new = (torch.stack(h_list, dim=0).unsqueeze(0).repeat(exer_seq.shape[0], 1, 1,1) *  cos_sim.unsqueeze(dim=-1).unsqueeze(dim=-1)).sum(dim=1)

            s = (beta.unsqueeze(-1)* h_new).sum(dim=1) # B*K, K*d -> B*d
            pred_v = torch.cat([exer_emb[:, t+1, :], s], dim=-1) # 1*B, 即每个学生在此刻对下一题的预测
            pd_t = self.pd_layer(pred_v)

            # 得到hidden vector
            xk = beta.unsqueeze(dim=-1) * fused_exer_emb[:,[t],:]
            _, h = self.stu_inter_seq_model(xk, h.unsqueeze(0))
            h = h.squeeze(0)
            h_list.append(h)
            pd_list.append(pd_t)

        return torch.concat(pd_list, dim=-1)



from ..common import EduDataTPL


class DKTForgetDataTPL(EduDataTPL):
    default_cfg = {
        'mid2cache_op_seq': ['M2C_Label2Int', 'M2C_ReMapId', 'M2C_BuildSeqInterFeats','M2C_RandomDataSplit4KT', "M2C_DKTForget_OP"],
        'M2C_BuildSeqInterFeats': {
            "extra_inter_feats": ['start_timestamp:float']
        }
    }

    def set_info_for_fold(self, fold_id):
        dt_info = self.datatpl_cfg['dt_info']
        dt_info['n_pcount'] = dt_info['n_pcount_list'][fold_id]
        dt_info['n_rgap'] = dt_info['n_rgap_list'][fold_id]
        dt_info['n_sgap'] = dt_info['n_sgap_list'][fold_id]



# -*- coding: utf-8 -*-
# @Author : Xiangzhi Chen
# @Github : kervias

from edustudio.utils.common import UnifyConfig, PathUtil, Logger
import os


def init_all(cfg: UnifyConfig):
    """initialize process

    Args:
        cfg (UnifyConfig): the global config obejct
    """
    frame_cfg = cfg.frame_cfg
    dataset = cfg.dataset
    traintpl_cls_name = cfg.traintpl_cfg.cls if isinstance(cfg.traintpl_cfg.cls, str) else cfg.traintpl_cfg.cls.__name__
    model_cls_name = cfg.modeltpl_cfg.cls if isinstance(cfg.modeltpl_cfg.cls, str) else cfg.modeltpl_cfg.cls.__name__

    frame_cfg.data_folder_path = f"{frame_cfg.DATA_FOLDER_PATH}/{dataset}"
    # PathUtil.check_path_exist(frame_cfg.data_folder_path)

    frame_cfg.TEMP_FOLDER_PATH = os.path.realpath(frame_cfg.TEMP_FOLDER_PATH)
    frame_cfg.ARCHIVE_FOLDER_PATH = os.path.realpath(frame_cfg.ARCHIVE_FOLDER_PATH)

    frame_cfg.temp_folder_path = f"{frame_cfg.TEMP_FOLDER_PATH}/{dataset}/{traintpl_cls_name}/{model_cls_name}/{frame_cfg.ID}"
    frame_cfg.archive_folder_path = f"{frame_cfg.ARCHIVE_FOLDER_PATH}/{dataset}/{traintpl_cls_name}/{model_cls_name}"
    PathUtil.auto_create_folder_path(
        frame_cfg.temp_folder_path, frame_cfg.archive_folder_path
    )
    log_filepath = f"{frame_cfg.temp_folder_path}/{frame_cfg.ID}.log"
    if frame_cfg['LOG_WITHOUT_DATE']:
        cfg.logger = Logger(
            filepath=log_filepath, fmt='[%(levelname)s]: %(message)s', date_fmt=None,
            DISABLE_LOG_STDOUT=cfg['frame_cfg']['DISABLE_LOG_STDOUT']
        ).get_std_logger()
    else:
        cfg.logger = Logger(
            filepath=log_filepath, DISABLE_LOG_STDOUT=cfg['frame_cfg']['DISABLE_LOG_STDOUT']
        ).get_std_logger()

    if frame_cfg['DISABLE_TQDM_BAR'] is True:
        from tqdm import tqdm
        from functools import partialmethod
        tqdm.__init__ = partialmethod(tqdm.__init__, disable=True)



import networkx as nx
import numpy as np

from scipy import sparse
from ..common import EduDataTPL


class RKTDataTPL(EduDataTPL):
    default_cfg = {
        'mid2cache_op_seq': ['M2C_Label2Int', 'M2C_ReMapId','M2C_GenQMat', 'M2C_BuildSeqInterFeats', 'M2C_RandomDataSplit4KT'],
        'M2C_BuildSeqInterFeats': {
            "extra_inter_feats": ['start_timestamp:float']
        }
    }

    def process_load_data_from_middata(self):
        super().process_load_data_from_middata()
        self.final_kwargs['pro_pro_dense'] = self.get_pro_pro_corr()

    def get_pro_pro_corr(self):
        # reference: https://github.com/shalini1194/RKT/issues/2
        pro_cpt_adj = []
        pro_num = self.cfg['datatpl_cfg']['dt_info']['exer_count']
        cpt_num = self.cfg['datatpl_cfg']['dt_info']['cpt_count']
        for index in range(len(self.df_exer)):
            tmp_df = self.df_exer.iloc[index]
            exer_id = tmp_df['exer_id:token']
            cpt_seq = tmp_df['cpt_seq:token_seq']
            for cpt in cpt_seq:
                pro_cpt_adj.append([exer_id, cpt, 1])
        pro_cpt_adj = np.array(pro_cpt_adj).astype(np.int32)
        pro_cpt_sparse = sparse.coo_matrix((pro_cpt_adj[:, 2].astype(np.float32),
                                              (pro_cpt_adj[:, 0], pro_cpt_adj[:, 1])), shape=(pro_num, cpt_num))
        pro_cpt_csc = pro_cpt_sparse.tocsc()
        pro_cpt_csr = pro_cpt_sparse.tocsr()
        pro_pro_adj = []
        for p in range(pro_num):
            tmp_skills = pro_cpt_csr.getrow(p).indices
            similar_pros = pro_cpt_csc[:, tmp_skills].indices
            zipped = zip([p] * similar_pros.shape[0], similar_pros)
            pro_pro_adj += list(zipped)

        pro_pro_adj = list(set(pro_pro_adj))
        pro_pro_adj = np.array(pro_pro_adj).astype(np.int32)
        data = np.ones(pro_pro_adj.shape[0]).astype(np.float32)
        pro_pro_sparse = sparse.coo_matrix((data, (pro_pro_adj[:, 0], pro_pro_adj[:, 1])), shape=(pro_num, pro_num))
        return 1-pro_pro_sparse.tocoo().toarray()

    def get_extra_data(self):
        return {
            "pro_pro_dense": self.final_kwargs['pro_pro_dense']
        }

    def set_info_for_fold(self, fold_id):
        super().set_info_for_fold(fold_id)
        self.train_dict = self.dict_train_folds[fold_id]


r"""
RKT
##########################################

Reference:
    Shalini Pandey et al. "RKT: Relation-Aware Self-Attention for Knowledge Tracing." in CIKM 2020.

Reference Code:
    https://github.com/shalini1194/RKT/blob/master/RKT/model_rkt.py

"""
import math
import copy
from ..gd_basemodel import GDBaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F
import numpy as np
from scipy import sparse
from torch.nn.init import xavier_uniform_, constant_

class RKT(GDBaseModel):
    r"""
   RKT

   default_cfg:
      'embed_size': 200  # dimension of embedding
      'drop_prob': 0.2      # dropout rate
      'num_attn_layers': 1        # number of attention layers
        'num_heads': 5         # number of parallel attention heads
        'encode_pos':False            # if True, use relative position embeddings
         'max_pos': 10          # number of position embeddings to use
   """
    default_cfg = {
        'embed_size': 200,
        'num_attn_layers': 1,
        'num_heads': 5,
        'encode_pos': False,
        'max_pos': 10,
        'drop_prob':0.2,
    }


    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        self.embed_size = self.modeltpl_cfg['embed_size']
        self.num_attn_layers = self.modeltpl_cfg['num_attn_layers']
        self.num_heads = self.modeltpl_cfg['num_heads']
        self.encode_pos = self.modeltpl_cfg['encode_pos']
        self.max_pos = self.modeltpl_cfg['max_pos']
        self.drop_prob = self.modeltpl_cfg['drop_prob']
        
    def build_model(self):
        self.item_embeds = nn.Embedding(self.n_item, self.embed_size, padding_idx=0)
        # self.skill_embeds = nn.Embedding(num_skills + 1, embed_size // 2, padding_idx=0)

        self.pos_key_embeds = nn.Embedding(self.max_pos, self.embed_size // self.num_heads)
        self.pos_value_embeds = nn.Embedding(self.max_pos, self.embed_size // self.num_heads)

        self.lin_in = nn.Linear(2 * self.embed_size, self.embed_size)
        self.attn_layers = clone(MultiHeadedAttention(self.embed_size, self.num_heads, self.drop_prob, self.device), self.num_attn_layers)
        self.dropout = nn.Dropout(p=self.drop_prob)
        self.lin_out = nn.Linear(self.embed_size, 1)
        self.l1 = nn.Parameter(torch.rand(1))
        self.l2 = nn.Parameter(torch.rand(1))
        # self.pro_pro_dense = self.get_corr_data()

    def get_inputs(self, item_inputs, label_inputs):
        item_inputs = self.item_embeds(item_inputs)
        # skill_inputs = self.skill_embeds(skill_inputs)
        label_inputs = label_inputs.unsqueeze(-1).float()

        inputs = torch.cat([item_inputs, item_inputs], dim=-1)
        inputs[..., :self.embed_size] *= label_inputs
        inputs[..., self.embed_size:] *= 1 - label_inputs
        return inputs

    def get_query(self, item_ids):
        item_ids = self.item_embeds(item_ids)
        # skill_ids = self.skill_embeds(skill_ids)
        query = torch.cat([item_ids], dim=-1)
        return query

    def add_extra_data(self, **kwargs):
        self.pro_pro_dense = kwargs['pro_pro_dense']

    def forward(self, exer_seq, label_seq,  start_timestamp_seq, **kwargs):
        time = computeRePos(start_timestamp_seq)
        item_inputs = torch.cat((torch.zeros((exer_seq.shape[0], 1)).to(self.device), exer_seq[:, :-1]), dim=1)
        label_inputs = torch.cat((torch.zeros((exer_seq.shape[0], 1)).to(self.device), label_seq[:, :-1]), dim=1)
        item_inputs = item_inputs.long().cpu()
        exer_seq = exer_seq.cpu()
        rel = self.pro_pro_dense[
            (exer_seq).unsqueeze(1).repeat(1, exer_seq.shape[-1], 1), (item_inputs).unsqueeze(-1).repeat(1, 1,
                                                                                                                 item_inputs.shape[
                                                                                                                 -1])]
        item_inputs = item_inputs.to(self.device)
        exer_seq = exer_seq.to(self.device)
        rel = torch.tensor(rel).to(self.device)

        inputs = self.get_inputs(item_inputs, label_inputs)

        inputs = F.relu(self.lin_in(inputs))

        query = self.get_query(exer_seq)

        mask = future_mask(inputs.size(-2))
        if inputs.is_cuda:
            mask = mask.to(self.device)
        outputs, attn = self.attn_layers[0](query, inputs, inputs, rel, self.l1, self.l2, time, self.encode_pos,
                                            self.pos_key_embeds, self.pos_value_embeds, mask)
        outputs = self.dropout(outputs)
        for l in self.attn_layers[1:]:
            residual, attn = l(query, outputs, outputs, rel, self.l1, self.l2, self.encode_pos, time,
                               self.pos_key_embeds,
                               self.pos_value_embeds, mask)
            outputs = self.dropout(outputs + F.relu(residual))

        return torch.sigmoid(self.lin_out(outputs))

    @torch.no_grad()
    def predict(self, **kwargs):
        y_pd = self(**kwargs)
        y_pd = y_pd[:, :-1]
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = None
        if kwargs.get('label_seq', None) is not None:
            y_gt = kwargs['label_seq'][:, 1:]
            y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        return {
            'y_pd': y_pd.squeeze(1),
            'y_gt': y_gt
        }

    def get_main_loss(self, **kwargs):
        y_pd = self(**kwargs)
        y_pd = y_pd[:, :-1]
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = kwargs['label_seq'][:, 1:]
        y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        loss = F.binary_cross_entropy(
            input=y_pd.squeeze(1), target=y_gt
        )
        return {
            'loss_main': loss,
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)



def future_mask(seq_length):
    future_mask = np.triu(np.ones((1, seq_length, seq_length)), k=1).astype('bool')
    return torch.from_numpy(future_mask)

def computeRePos(time_seq):
    batch_size = time_seq.shape[0]
    size = time_seq.shape[1]

    time_matrix= (torch.abs(torch.unsqueeze(time_seq, axis=1).repeat(1,size,1).reshape((batch_size, size*size,1)) - \
                 torch.unsqueeze(time_seq,axis=-1).repeat(1, 1, size,).reshape((batch_size, size*size,1))))

    # time_matrix[time_matrix>time_span] = time_span
    time_matrix = time_matrix.reshape((batch_size,size,size))


    return (time_matrix)

def clone(module, num):
    return nn.ModuleList([copy.deepcopy(module) for _ in range(num)])


def attention(query, key, value, rel, l1, l2, timestamp, mask=None, dropout=None, device='cpu'):
    """Compute scaled dot product attention.
    """
    rel = rel * mask.to(torch.float) # future masking of correlation matrix.
    rel_attn = rel.masked_fill(rel == 0, -10000)
    rel_attn = nn.Softmax(dim=-1)(rel_attn)
    scores = torch.matmul(query, key.transpose(-2, -1))
    scores = scores / math.sqrt(query.size(-1))
    if mask is not None:
        scores = scores.masked_fill(mask, -1e9)

        time_stamp= torch.exp(-torch.abs(timestamp.float()))
        #
        time_stamp=time_stamp.masked_fill(mask,-np.inf)


    prob_attn = F.softmax(scores, dim=-1)
    time_attn = F.softmax(time_stamp,dim=-1)
    prob_attn = (1-l2)*prob_attn+l2*time_attn
    # prob_attn = F.softmax(prob_attn + rel_attn, dim=-1)

    prob_attn = (1-l1)*prob_attn + (l1)*rel_attn
    if dropout is not None:
        prob_attn = dropout(prob_attn)
    return torch.matmul(prob_attn, value), prob_attn


def relative_attention(query, key, value, rel, l1, l2, pos_key_embeds, pos_value_embeds, mask=None, dropout=None, device='cpu'):
    """Compute scaled dot product attention with relative position embeddings.
    (https://arxiv.org/pdf/1803.02155.pdf)
    """
    assert pos_key_embeds.num_embeddings == pos_value_embeds.num_embeddings

    scores = torch.matmul(query, key.transpose(-2, -1))

    idxs = torch.arange(scores.size(-1))
    if query.is_cuda:
        idxs = idxs.to(device)
    idxs = idxs.view(-1, 1) - idxs.view(1, -1)
    idxs = torch.clamp(idxs, 0, pos_key_embeds.num_embeddings - 1)

    pos_key = pos_key_embeds(idxs).transpose(-2, -1)
    pos_scores = torch.matmul(query.unsqueeze(-2), pos_key)
    scores = scores.unsqueeze(-2) + pos_scores
    scores = scores / math.sqrt(query.size(-1))

    pos_value = pos_value_embeds(idxs)
    value = value.unsqueeze(-3) + pos_value

    if mask is not None:
        scores = scores.masked_fill(mask.unsqueeze(-2), -1e9)
    prob_attn = F.softmax(scores, dim=-1)
    if dropout is not None:
        prob_attn = dropout(prob_attn)

    output = torch.matmul(prob_attn, value).unsqueeze(-2)
    prob_attn = prob_attn.unsqueeze(-2)
    return output, prob_attn


class MultiHeadedAttention(nn.Module):
    def __init__(self, total_size, num_heads, drop_prob, device):
        super(MultiHeadedAttention, self).__init__()
        assert total_size % num_heads == 0
        self.device = device
        self.total_size = total_size
        self.head_size = total_size // num_heads
        self.num_heads = num_heads
        self.linear_layers = clone(nn.Linear(total_size, total_size), 3)
        self.dropout = nn.Dropout(p=drop_prob)

    def forward(self, query, key, value, rel, l1, l2, timestamp, encode_pos, pos_key_embeds, pos_value_embeds, mask=None):
        batch_size, seq_length = query.shape[:2]

        # Apply mask to all heads
        if mask is not None:
            mask = mask.unsqueeze(1)

        # Project inputs
        rel = rel.unsqueeze(1).repeat(1,self.num_heads,1,1)
        timestamp = timestamp.unsqueeze(1).repeat(1,self.num_heads,1,1)
        query, key, value = [l(x).view(batch_size, seq_length, self.num_heads, self.head_size).transpose(1, 2)
                             for l, x in zip(self.linear_layers, (query, key, value))]

        # Apply attention
        if encode_pos:
            out, self.prob_attn = relative_attention(
                query, key, value, rel, l1, l2, timestamp, pos_key_embeds, pos_value_embeds,  mask, self.dropout, self.device)
        else:
            out, self.prob_attn = attention(query, key, value, rel, l1, l2, timestamp, mask, self.dropout, self.device)

        out = out.transpose(1, 2).contiguous().view(batch_size, seq_length, self.total_size)
        return out, self.prob_attn




from ..gd_basemodel import GDBaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F


class QIKT(GDBaseModel):
    default_cfg = {
        'emb_size': 100,
        'hidden_size': 100,
        'mlp_layer_num': 1,
        'dropout_rate': 0.2,
        'output_c_all_lambda': 1.0,
        'output_c_next_lambda': 1.0,
        'output_q_all_lambda': 1.0,
        'output_q_next_lambda': 0.0,
        'loss_c_all_lambda': 0.0,
        'loss_c_next_lambda': 0.0,
        'loss_q_all_lambda': 0.0,
        'loss_q_next_lambda': 0.0,
        'output_mode': 'an_irt',
    }

    def __init__(self, cfg):
        super().__init__(cfg)

    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        self.n_cpt = self.datatpl_cfg['dt_info']['cpt_count']
        self.output_c_all_lambda = self.modeltpl_cfg['output_c_all_lambda']
        self.output_c_next_lambda = self.modeltpl_cfg['output_c_next_lambda']
        self.output_q_all_lambda = self.modeltpl_cfg['output_q_all_lambda']
        self.output_q_next_lambda = self.modeltpl_cfg['output_q_next_lambda']
        self.loss_c_all_lambda = self.modeltpl_cfg['loss_c_all_lambda']
        self.loss_c_next_lambda = self.modeltpl_cfg['loss_c_next_lambda']
        self.loss_q_all_lambda = self.modeltpl_cfg['loss_q_all_lambda']
        self.loss_q_next_lambda = self.modeltpl_cfg['loss_q_next_lambda']
        self.output_mode = self.modeltpl_cfg['output_mode']
        self.device = self.traintpl_cfg['device']

    def build_model(self):
        num_q, num_c = self.n_item, self.n_cpt
        self.exer_emb = nn.Embedding(
            self.n_item, self.modeltpl_cfg['emb_size']
        )
        self.cpt_emb = nn.Embedding(
            self.n_cpt, self.modeltpl_cfg['emb_size']
        )
        self.que_lstm_layer = nn.LSTM(self.modeltpl_cfg['emb_size'] * 4, self.modeltpl_cfg['hidden_size'], batch_first=True)
        self.concept_lstm_layer = nn.LSTM(self.modeltpl_cfg['emb_size'] * 2, self.modeltpl_cfg['hidden_size'], batch_first=True)

        self.dropout_layer = nn.Dropout(self.modeltpl_cfg['dropout_rate'])

        self.out_question_next = MLP(self.modeltpl_cfg['mlp_layer_num'], self.modeltpl_cfg['hidden_size'] * 3, 1, self.modeltpl_cfg['dropout_rate'])
        self.out_question_all = MLP(self.modeltpl_cfg['mlp_layer_num'], self.modeltpl_cfg['hidden_size'], num_q, self.modeltpl_cfg['dropout_rate'])

        self.out_concept_next = MLP(self.modeltpl_cfg['mlp_layer_num'], self.modeltpl_cfg['hidden_size'] * 3, num_c, self.modeltpl_cfg['dropout_rate'])
        self.out_concept_all = MLP(self.modeltpl_cfg['mlp_layer_num'], self.modeltpl_cfg['hidden_size'], num_c, self.modeltpl_cfg['dropout_rate'])

        self.que_disc = MLP(self.modeltpl_cfg['mlp_layer_num'], self.modeltpl_cfg['hidden_size'] * 2, 1, self.modeltpl_cfg['dropout_rate'])

    def forward(self, exer_seq, label_seq, cpt_seq, cpt_seq_mask, **kwargs):



        # obtain emb_q, emb_c, emb_qca, emb_qc
        emb_q = self.exer_emb(exer_seq)
        k = self.cpt_emb(cpt_seq)
        emb_c = torch.sum(k * (cpt_seq_mask.unsqueeze(3).repeat(1, 1, 1, self.modeltpl_cfg['emb_size'])),
                          dim=2) / cpt_seq_mask.sum(dim=2).unsqueeze(2).repeat(1, 1, self.modeltpl_cfg['emb_size'])


        emb_qc = torch.cat((emb_q, emb_c), dim=2)
        mask_e = label_seq.unsqueeze(-1).repeat(1, 1, emb_qc.shape[-1]).to(torch.float)
        emb_qca = torch.cat((mask_e*emb_qc, (1-mask_e)*emb_qc), dim=-1)

        emb_qc_shift = emb_qc[:, 1:, :]
        emb_qca_current = emb_qca[:, :-1, :]
        # question model
        que_h = self.dropout_layer(self.que_lstm_layer(emb_qca_current)[0])
        que_outputs = self.get_outputs(emb_qc_shift, que_h, data=exer_seq[:, 1:], add_name="", modeltpl_type="question")
        outputs = que_outputs

        # concept model
        emb_ca = torch.cat([emb_c.mul((1 - label_seq).unsqueeze(-1).repeat(1, 1, self.modeltpl_cfg['emb_size'])),
                            emb_c.mul((label_seq).unsqueeze(-1).repeat(1, 1, self.modeltpl_cfg['emb_size']))], dim=-1)  # s_t 扩展，分别对应正确的错误的情况

        emb_ca_current = emb_ca[:, :-1, :]
        # emb_c_shift = emb_c[:,1:,:]
        concept_h = self.dropout_layer(self.concept_lstm_layer(emb_ca_current)[0])
        concept_outputs = self.get_outputs(emb_qc_shift, concept_h, data=(cpt_seq[:, 1:, :], cpt_seq_mask[:, 1:, :]), add_name="", modeltpl_type="concept")
        outputs['y_concept_all'] = concept_outputs['y_concept_all']
        outputs['y_concept_next'] = concept_outputs['y_concept_next']


        if self.output_mode=="an_irt":
            def sigmoid_inverse(x,epsilon=1e-8):
                return torch.log(x/(1-x+epsilon)+epsilon)
            y = sigmoid_inverse(outputs['y_question_all'])*self.output_q_all_lambda + \
                sigmoid_inverse(outputs['y_concept_all'])*self.output_c_all_lambda + \
                sigmoid_inverse(outputs['y_concept_next'])*self.output_c_next_lambda
            y = torch.sigmoid(y)
        else:
            # output weight
            y = outputs['y_question_all'] * self.output_q_all_lambda \
                + outputs['y_concept_all'] * self.output_c_all_lambda \
                + outputs['y_concept_next'] * self.output_c_next_lambda
            y = y/(self.output_q_all_lambda + self.output_c_all_lambda + self.output_c_next_lambda)
        outputs['y'] = y
        return outputs

    @torch.no_grad()
    def predict(self, **kwargs):
        outputs = self(**kwargs)
        y_pd = outputs['y']
        # y_pd = y_pd[:, :-1].gather(
        #     index=kwargs['exer_seq'][:, 1:].unsqueeze(dim=-1), dim=2
        # ).squeeze(dim=-1)
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = None
        if kwargs.get('label_seq', None) is not None:
            y_gt = kwargs['label_seq'][:, 1:]
            y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        return {
            'y_pd': y_pd,
            'y_gt': y_gt
        }

    def get_main_loss(self, **kwargs):
        outputs = self(**kwargs)

        rshft = kwargs['label_seq'][:, 1:]
        maskshft = kwargs['mask_seq'][:, 1:].bool()
        # all
        loss_q_all = self.get_loss(outputs['y_question_all'],rshft,maskshft)
        loss_c_all = self.get_loss(outputs['y_concept_all'],rshft,maskshft)
        # next
        loss_q_next = self.get_loss(outputs['y_question_next'],rshft,maskshft)#question level loss
        loss_c_next = self.get_loss(outputs['y_concept_next'],rshft,maskshft)#kc level loss
        # over all
        loss_kt = self.get_loss(outputs['y'],rshft,maskshft)

        if self.output_mode == "an_irt":
            loss = loss_kt + \
                   self.loss_q_all_lambda * loss_q_all + \
                   self.loss_c_all_lambda * loss_c_all + \
                   self.loss_c_next_lambda * loss_c_next
        else:
            loss = loss_kt + \
                   self.loss_q_all_lambda * loss_q_all + \
                   self.loss_c_all_lambda * loss_c_all + \
                   self.loss_c_next_lambda * loss_c_next + \
                   self.loss_q_next_lambda * loss_q_next

        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)


    def get_outputs(self, emb_qc_shift, h, data, add_name="", modeltpl_type='question'):
        outputs = {}

        if modeltpl_type == 'question':
            h_next = torch.cat([emb_qc_shift, h], axis=-1)
            y_question_next = torch.sigmoid(self.out_question_next(h_next))
            y_question_all = torch.sigmoid(self.out_question_all(h))
            outputs["y_question_next" + add_name] = y_question_next.squeeze(-1)
            outputs["y_question_all" + add_name] = (y_question_all * F.one_hot(data.long(), self.n_item)).sum(-1)
        else:
            h_next = torch.cat([emb_qc_shift, h], axis=-1)
            y_concept_next = torch.sigmoid(self.out_concept_next(h_next))
            # all predict
            y_concept_all = torch.sigmoid(self.out_concept_all(h))
            outputs["y_concept_next" + add_name] = self.get_avg_fusion_concepts(y_concept_next, data)
            outputs["y_concept_all" + add_name] = self.get_avg_fusion_concepts(y_concept_all, data)

        return outputs

    def get_avg_fusion_concepts(self, y_concept, cshft):
        max_num_concept = cshft[0].shape[-1]
        # concept_mask = torch.where(cshft.long()==-1, False, True)
        concept_index = F.one_hot(cshft[0], self.n_cpt)
        # concept_index = F.one_hot(torch.where(cshft!=-1,cshft,0), self.num_c)
        concept_sum = (y_concept.unsqueeze(2).repeat(1,1,max_num_concept,1)*concept_index).sum(-1)
        concept_sum = concept_sum*cshft[1]#remove mask
        y_concept = concept_sum.sum(-1)/torch.where(cshft[1].sum(-1)!=0,cshft[1].sum(-1),1)
        return y_concept

    def get_loss(self, ys,rshft,sm):
        y_pred = torch.masked_select(ys, sm)
        y_true = torch.masked_select(rshft, sm)
        loss = F.binary_cross_entropy(
            input=y_pred.double(), target=y_true.double()
        )
        return loss

class MLP(nn.Module):
    def __init__(self, n_layer, hidden_dim, output_dim, dpo):
        super().__init__()

        self.lins = nn.ModuleList([
            nn.Linear(hidden_dim, hidden_dim)
            for _ in range(n_layer)
        ])
        self.dropout = nn.Dropout(p = dpo)
        self.out = nn.Linear(hidden_dim, output_dim)
        self.act = torch.nn.Sigmoid()

    def forward(self, x):
        for lin in self.lins:
            x = F.relu(lin(x))
        return self.out(self.dropout(x))


from .general_traintpl import GeneralTrainTPL
import torch
from collections import defaultdict
import numpy as np
from tqdm import tqdm


class AdversarialTrainTPL(GeneralTrainTPL):
    default_cfg = {
        'lr': 0.001,
        'lr_d': 0.001,
        'g_rounds': 1,
        'd_rounds': 1,
        'optim': 'adam',
        'optim_d': 'adam',
    }

    def _get_optim(self, model_params, optimizer='adam', lr=0.001, weight_decay=0.0, eps=1e-8):
        """Get optimizer
        """
        if optimizer == "sgd":
            optim = torch.optim.SGD(model_params, lr=lr, weight_decay=weight_decay, eps=eps)
        elif optimizer == "adam":
            optim = torch.optim.Adam(model_params, lr=lr, weight_decay=weight_decay, eps=eps)
        elif optimizer == "adagrad":
            optim = torch.optim.Adagrad(model_params, lr=lr, weight_decay=weight_decay, eps=eps)
        elif optimizer == "rmsprop":
            optim = torch.optim.RMSprop(model_params, lr=lr, weight_decay=weight_decay, eps=eps)
        else:
            raise ValueError("unsupported optimizer")

        return optim

    def fit(self, train_loader, valid_loader):
        self.model.train()
        lr = self.traintpl_cfg['lr']
        lr_d = self.traintpl_cfg['lr_d']
        weight_decay = self.traintpl_cfg['weight_decay']
        eps = self.traintpl_cfg['eps']
        
        self.optimizer_g = self._get_optim(self.model.get_g_parameters(), self.traintpl_cfg['optim'], lr=lr, weight_decay=weight_decay, eps=eps)
        self.optimizer_d = self._get_optim(self.model.get_d_parameters(), self.traintpl_cfg['optim_d'], lr=lr_d, weight_decay=weight_decay, eps=eps)

        self.callback_list.on_train_begin()
        for epoch in range(self.traintpl_cfg['epoch_num']):
            self.callback_list.on_epoch_begin(epoch + 1)

            # train_for_generator
            g_rounds = self.traintpl_cfg['g_rounds']
            d_rounds = self.traintpl_cfg['d_rounds']
            logs = defaultdict(lambda: np.full((len(train_loader) * g_rounds,), np.nan, dtype=np.float32))
            for round_id in range(g_rounds):
                for batch_id, batch_dict in enumerate(
                        tqdm(train_loader, ncols=self.frame_cfg['TQDM_NCOLS'], desc="[GEN:EPOCH={:03d}]".format(epoch + 1))
                ):
                    batch_dict = self.batch_dict2device(batch_dict)
                    loss_gen_dict = self.model.get_main_loss(**batch_dict)
                    loss_gen = torch.hstack([i for i in loss_gen_dict.values() if i is not None]).sum()
                    loss = loss_gen
                    self.optimizer_g.zero_grad()
                    loss.backward()
                    self.optimizer_g.step()
                    for k in loss_gen_dict: logs[k][batch_id + len(train_loader) * round_id] = loss_gen_dict[k].item() if loss_gen_dict[k] is not None else np.nan

            logs_g = {}
            for name in logs: logs_g[f"GEN_{name}"] = float(np.nanmean(logs[name]))

            # train_for_discriminator
            logs = defaultdict(lambda: np.full((len(train_loader) * d_rounds,), np.nan, dtype=np.float32))
            for round_id in range(d_rounds):
                for batch_id, batch_dict in enumerate(
                        tqdm(train_loader, ncols=self.frame_cfg['TQDM_NCOLS'], desc="[DIS:EPOCH={:03d}]".format(epoch + 1))
                ):
                    batch_dict = self.batch_dict2device(batch_dict)
                    loss_pre_dict, loss_dis_dict = self.model.get_loss_dict(**batch_dict)
                    loss_pre = torch.hstack([i for i in loss_pre_dict.values() if i is not None]).sum()
                    loss_dis = torch.hstack([i for i in loss_dis_dict.values() if i is not None]).sum()
                    loss = loss_pre - loss_dis
                    self.optimizer_d.zero_grad()
                    loss.backward()
                    self.optimizer_d.step()
                    for k in loss_pre_dict: logs[k][batch_id + len(train_loader) * round_id] = loss_pre_dict[k].item() if loss_pre_dict[k] is not None else np.nan
                    for k in loss_dis_dict: logs[k][batch_id + len(train_loader) * round_id] = loss_dis_dict[k].item() if loss_dis_dict[k] is not None else np.nan

            logs_d = {}
            for name in logs: logs_d[f"DIS_{name}"] = float(np.nanmean(logs[name]))

            logs = logs_g
            logs.update(logs_d)
            if valid_loader is not None:
                val_metrics = self.evaluate(valid_loader)
                logs.update({f"{metric}": val_metrics[metric] for metric in val_metrics})

            self.callback_list.on_epoch_end(epoch + 1, logs=logs)
            if self.model.share_callback_dict.get('stop_training', False):
                break
            
        self.callback_list.on_train_end()



from ..common import EduDataTPL
import numpy as np


class KTInterCptUnfoldDataTPL(EduDataTPL):
    default_cfg = {
        'mid2cache_op_seq': ['M2C_Label2Int', 'M2C_ReMapId', 'M2C_GenUnFoldKCSeq', 'M2C_BuildSeqInterFeats', 'M2C_RandomDataSplit4KT'],
        'M2C_BuildSeqInterFeats': {
            "extra_inter_feats": ['start_timestamp:float', 'cpt_unfold:token']
        }
    }



r"""
RCD
##################################
Reference:
    Weibo Gao et al. "RCD: Relation Map Driven Cognitive Diagnosis for Intelligent Education Systems." in SIGIR 2021.

Reference Code:
    https://github.com/bigdata-ustc/RCD
"""


from ..gd_basemodel import GDBaseModel
import torch.nn as nn
from ..utils.components import PosMLP
import torch
import torch.nn.functional as F


class GraphLayer(nn.Module):
    def __init__(self, g, in_dim, out_dim):
        super(GraphLayer, self).__init__()
        self.g = g
        self.fc = nn.Linear(in_dim, out_dim, bias=False)
        self.attn_fc = nn.Linear(2 * out_dim, 1, bias=False)

    def edge_attention(self, edges):
        z2 = torch.cat([edges.src['z'], edges.dst['z']], dim=1)
        a = self.attn_fc(z2)
        return {'e': a}

    def message_func(self, edges):
        return {'z': edges.src['z'], 'e': edges.data['e']}

    def reduce_func(self, nodes):
        alpha = F.softmax(nodes.mailbox['e'], dim=1)
        h = torch.sum(alpha * nodes.mailbox['z'], dim=1)
        return {'h': h}

    def forward(self, h):
        z = self.fc(h)
        self.g.ndata['z'] = z
        self.g.apply_edges(self.edge_attention)
        self.g.update_all(self.message_func, self.reduce_func)
        return self.g.ndata.pop('h')


class Fusion(nn.Module):
    def __init__(self, n_user, n_item, n_cpt, local_map, device):
        self.device = device
        self.knowledge_dim = n_cpt
        self.exer_n = n_item
        self.emb_num = n_user
        self.stu_dim = self.knowledge_dim

        # graph structure
        self.directed_g = local_map['directed_g'].to(self.device)
        self.undirected_g = local_map['undirected_g'].to(self.device)
        self.k_from_e = local_map['k_from_e'].to(self.device)
        self.e_from_k = local_map['e_from_k'].to(self.device)
        self.u_from_e = local_map['u_from_e'].to(self.device)
        self.e_from_u = local_map['e_from_u'].to(self.device)

        super(Fusion, self).__init__()

        self.directed_gat = GraphLayer(self.directed_g, n_cpt, n_cpt)
        self.undirected_gat = GraphLayer(self.undirected_g, n_cpt, n_cpt)

        self.k_from_e = GraphLayer(self.k_from_e, n_cpt, n_cpt)  # src: e
        self.e_from_k = GraphLayer(self.e_from_k, n_cpt, n_cpt)  # src: k

        self.u_from_e = GraphLayer(self.u_from_e, n_cpt, n_cpt)  # src: e
        self.e_from_u = GraphLayer(self.e_from_u, n_cpt, n_cpt)  # src: u

        self.k_attn_fc1 = nn.Linear(2 * n_cpt, 1, bias=True)
        self.k_attn_fc2 = nn.Linear(2 * n_cpt, 1, bias=True)
        self.k_attn_fc3 = nn.Linear(2 * n_cpt, 1, bias=True)

        self.e_attn_fc1 = nn.Linear(2 * n_cpt, 1, bias=True)
        self.e_attn_fc2 = nn.Linear(2 * n_cpt, 1, bias=True)

    def forward(self, kn_emb, exer_emb, all_stu_emb):
        k_directed = self.directed_gat(kn_emb)
        k_undirected = self.undirected_gat(kn_emb)

        e_k_graph = torch.cat((exer_emb, kn_emb), dim=0)
        k_from_e_graph = self.k_from_e(e_k_graph)
        e_from_k_graph = self.e_from_k(e_k_graph)

        e_u_graph = torch.cat((exer_emb, all_stu_emb), dim=0)
        u_from_e_graph = self.u_from_e(e_u_graph)
        e_from_u_graph = self.e_from_u(e_u_graph)

        # update concepts
        A = kn_emb
        B = k_directed
        C = k_undirected
        D = k_from_e_graph[self.exer_n:]
        concat_c_1 = torch.cat([A, B], dim=1)
        concat_c_2 = torch.cat([A, C], dim=1)
        concat_c_3 = torch.cat([A, D], dim=1)
        score1 = self.k_attn_fc1(concat_c_1)
        score2 = self.k_attn_fc2(concat_c_2)
        score3 = self.k_attn_fc3(concat_c_3)
        score = F.softmax(torch.cat([torch.cat([score1, score2], dim=1), score3], dim=1),
                          dim=1)  # dim = 1, 按行SoftMax, 行和为1
        kn_emb = A + score[:, 0].unsqueeze(1) * B + score[:, 1].unsqueeze(1) * C + score[:, 2].unsqueeze(1) * D

        # updated exercises
        A = exer_emb
        B = e_from_k_graph[0: self.exer_n]
        C = e_from_u_graph[0: self.exer_n]
        concat_e_1 = torch.cat([A, B], dim=1)
        concat_e_2 = torch.cat([A, C], dim=1)
        score1 = self.e_attn_fc1(concat_e_1)
        score2 = self.e_attn_fc2(concat_e_2)
        score = F.softmax(torch.cat([score1, score2], dim=1), dim=1)  # dim = 1, 按行SoftMax, 行和为1
        exer_emb = exer_emb + score[:, 0].unsqueeze(1) * B + score[:, 1].unsqueeze(1) * C

        # updated students
        all_stu_emb = all_stu_emb + u_from_e_graph[self.exer_n:]

        return kn_emb, exer_emb, all_stu_emb


class RCD(GDBaseModel):
    default_cfg = {
        'prednet_len1': 512,
        'prednet_len2': 256
    }
    def __init__(self, cfg):
        super().__init__(cfg)

    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        self.n_cpt = self.datatpl_cfg['dt_info']['cpt_count']

    def add_extra_data(self, **kwargs):
        self.local_map = kwargs.pop('local_map')
        self.Q_mat = kwargs['Q_mat'].to(self.device)

    def build_model(self):
        self.knowledge_dim = self.n_cpt
        self.exer_n = self.n_item
        self.emb_num = self.n_user
        self.stu_dim = self.knowledge_dim
        self.prednet_input_len = self.knowledge_dim
        self.prednet_len1, self.prednet_len2 = self.modeltpl_cfg['prednet_len1'], self.modeltpl_cfg['prednet_len2']
        self.directed_g = self.local_map['directed_g'].to(self.device)
        self.undirected_g = self.local_map['undirected_g'].to(self.device)
        self.k_from_e = self.local_map['k_from_e'].to(self.device)
        self.e_from_k = self.local_map['e_from_k'].to(self.device)
        self.u_from_e = self.local_map['u_from_e'].to(self.device)
        self.e_from_u = self.local_map['e_from_u'].to(self.device)

        self.k_index = torch.LongTensor(list(range(self.stu_dim))).to(self.device)
        self.stu_index = torch.LongTensor(list(range(self.emb_num))).to(self.device)
        self.exer_index = torch.LongTensor(list(range(self.exer_n))).to(self.device)

        self.student_emb = nn.Embedding(self.n_user, self.n_cpt)
        self.knowledge_emb = nn.Embedding(self.n_cpt, self.n_cpt)
        self.exercise_emb = nn.Embedding(self.n_item, self.n_cpt)

        self.FusionLayer1 = Fusion(self.n_user, self.n_item, self.n_cpt, self.local_map, self.device)
        self.FusionLayer2 = Fusion(self.n_user, self.n_item, self.n_cpt, self.local_map, self.device)

        self.prednet_full1 = nn.Linear(2 * self.n_cpt, self.n_cpt, bias=False)
        self.prednet_full2 = nn.Linear(2 * self.n_cpt, self.n_cpt, bias=False)
        self.prednet_full3 = nn.Linear(1 * self.n_cpt, 1)


    def forward(self, stu_id, exer_id, **kwargs):
        # before prednet
        all_stu_emb = self.student_emb(self.stu_index).to(self.device)
        exer_emb = self.exercise_emb(self.exer_index).to(self.device)
        kn_emb = self.knowledge_emb(self.k_index).to(self.device)

        kn_emb1, exer_emb1, all_stu_emb1 = self.FusionLayer1(kn_emb, exer_emb, all_stu_emb)
        # Fusion layer 2
        kn_emb2, exer_emb2, all_stu_emb2 = self.FusionLayer2(kn_emb1, exer_emb1, all_stu_emb1)

        # get batch student data
        batch_stu_emb = all_stu_emb2[stu_id]  # 32 123
        batch_stu_vector = batch_stu_emb.repeat(1, batch_stu_emb.shape[1]).reshape(batch_stu_emb.shape[0],
                                                                                   batch_stu_emb.shape[1],
                                                                                   batch_stu_emb.shape[1])

        # get batch exercise data
        batch_exer_emb = exer_emb2[exer_id]  # 32 123
        batch_exer_vector = batch_exer_emb.repeat(1, batch_exer_emb.shape[1]).reshape(batch_exer_emb.shape[0],
                                                                                      batch_exer_emb.shape[1],
                                                                                      batch_exer_emb.shape[1])

        # get batch knowledge concept data
        kn_vector = kn_emb2.repeat(batch_stu_emb.shape[0], 1).reshape(batch_stu_emb.shape[0], kn_emb2.shape[0],
                                                                      kn_emb2.shape[1])

        # Cognitive diagnosis
        preference = torch.sigmoid(self.prednet_full1(torch.cat((batch_stu_vector, kn_vector), dim=2)))
        diff = torch.sigmoid(self.prednet_full2(torch.cat((batch_exer_vector, kn_vector), dim=2)))
        o = torch.sigmoid(self.prednet_full3(preference - diff))

        sum_out = torch.sum(o * self.Q_mat[exer_id].unsqueeze(2), dim=1)
        count_of_concept = torch.sum(self.Q_mat[exer_id], dim=1).unsqueeze(1)
        pd = sum_out / count_of_concept
        return pd

    @torch.no_grad()
    def predict(self, stu_id, exer_id, **kwargs):
        return {
            'y_pd': self(stu_id, exer_id).flatten(),
        }

    def get_main_loss(self, **kwargs):
        stu_id = kwargs['stu_id']
        exer_id = kwargs['exer_id']
        label = kwargs['label']
        pd = self(stu_id, exer_id).flatten()

        # output_1 = self(stu_id, exer_id, Q_mat)
        # output_0 = torch.ones(output_1.size()).to(self.device) - output_1
        # output = torch.cat((output_0, output_1), 1)
        # pd = torch.log(output+1e-10).flatten()

        loss = F.binary_cross_entropy(input=pd, target=label)
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)
    
    # def get_stu_status(self, stu_id=None):
    #     if stu_id is not None:
    #         return self.student_emb(stu_id)
    #     else:
    #         return self.student_emb.weight



r"""
CNCD-Q
##########################################

Reference:
    Fei Wang et al. "Neural Cognitive Diagnosis for Intelligent Education Systems" in AAAI 2020.

Reference Code:
    https://github.com/bigdata-ustc/Neural_Cognitive_Diagnosis-NeuralCD
    https://github.com/LegionKing/NeuralCDM_plus

"""
import numpy as np
from ..gd_basemodel import GDBaseModel
import torch.nn as nn
from ..utils.components import PosMLP
import torch
import torch.nn.functional as F


class CNCD_Q(GDBaseModel):
    r"""
    CNCD-Q

    default_cfg:
       'dnn_units': [512, 256]  # dimension list of hidden layer in prediction layer
       'dropout_rate': 0.5      # dropout rate
       'activation': 'sigmoid'  # activation function in prediction layer
       'disc_scale': 10         # discrimination scale
    """
    default_cfg = {
        'dnn_units': [512, 256],
        'dropout_rate': 0.5,
        'activation': 'sigmoid',
        'disc_scale': 10
    }
    def __init__(self, cfg):
        super().__init__(cfg)

    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        self.n_cpt = self.datatpl_cfg['dt_info']['cpt_count']

    def build_model(self):
        # prediction sub-net
        self.student_emb = nn.Embedding(self.n_user, self.n_cpt)
        self.k_difficulty = nn.Embedding(self.n_item, self.n_cpt)
        self.e_difficulty = nn.Embedding(self.n_item, 1)
        self.pd_net = PosMLP(
            input_dim=self.n_cpt, output_dim=1, activation=self.modeltpl_cfg['activation'],
            dnn_units=self.modeltpl_cfg['dnn_units'], dropout_rate=self.modeltpl_cfg['dropout_rate']
        )
        self.e_k_prob = nn.Embedding(self.n_item, self.n_cpt)

    def forward(self, stu_id, exer_id, Q_mat, **kwargs):
        # before prednet
        users = stu_id
        items = exer_id
        items_Q_mat = Q_mat[items]
        stu_emb = self.student_emb(users)
        stat_emb = torch.sigmoid(stu_emb)
        k_difficulty = torch.sigmoid(self.k_difficulty(items))
        e_difficulty = torch.sigmoid(self.e_difficulty(items)) * self.modeltpl_cfg['disc_scale']
        # prednet
        e_k_prob = self.e_k_prob(items)
        e_k_prob_2 = F.sigmoid(e_k_prob)  # knowledge relevancy vectors of the exercises
        input_x = e_difficulty * (stat_emb - k_difficulty) * (items_Q_mat * e_k_prob_2)
        pd = self.pd_net(input_x).sigmoid()
        if self.training:
            return pd, e_k_prob
        else:
            return pd

    @torch.no_grad()
    def predict(self, stu_id, exer_id, Q_mat, **kwargs):
        return {
            'y_pd': self(stu_id, exer_id, Q_mat).flatten(),
        }

    def get_main_loss(self, **kwargs):
        normal_mean, normal_C = 0, 2
        means = torch.ones(self.n_cpt) * normal_mean  # the mean of the multidimensional gaussian distribution
        means.require_grad = False
        means = means.to(self.traintpl_cfg['device'])
        C = torch.ones(self.n_cpt) * normal_C  # the diagonal of the covariance matrix
        C.require_grad = False
        C = C.to(self.traintpl_cfg['device'])
        stu_id = kwargs['stu_id']
        exer_id = kwargs['exer_id']
        label = kwargs['label']
        Q_mat = kwargs['Q_mat']
        knowledge_pairs_kw = kwargs['knowledge_pairs']
        # kn_tops = kwargs["kn_tops"]
        # kn_tags = kwargs["kn_tags"]
        pd, exer_knowledge_prob= self(stu_id, exer_id, Q_mat=Q_mat)
        pd = pd.flatten()
        loss_1 = F.binary_cross_entropy(input=pd, target=label)
        loss_2 = 0
        knowledge_pairs =[]
        for i in exer_id.cpu().numpy():
            knowledge_pairs.append(knowledge_pairs_kw[i])
        for pair_i in range(len(knowledge_pairs)):#batch_size
            kn_tags, kn_topks = knowledge_pairs[pair_i]
            kn_tags, kn_topks = np.array(kn_tags) - 0, np.array(kn_topks) - 0  # 转成从0开始
            # if -1 in kn_tags or -1 in kn_topks:
            #     print(kn_tags)
            kn_tag_n = len(kn_tags)
            kn_tag_tensor = exer_knowledge_prob[pair_i][kn_tags].view(-1, 1)
            kn_prob_tensor = exer_knowledge_prob[pair_i][kn_topks].repeat(kn_tag_n, 1)
            loss_2 = loss_2 - (torch.log(torch.sigmoid((kn_tag_tensor - kn_prob_tensor) * 0.1))).sum()
        for kn_prob in exer_knowledge_prob:
            a = kn_prob - means
            loss_2 = loss_2 + 0.5 * (a * a / C).sum()


        return {
            'loss_1': loss_1,
            'loss_2': loss_2
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)
    
    def get_stu_status(self, stu_id=None):
        if stu_id is not None:
            return self.student_emb(stu_id)
        else:
            return self.student_emb.weight



from edustudio.utils.common import UnifyConfig
from .base_mid2cache import BaseMid2Cache
from sklearn.preprocessing import LabelEncoder
import pandas as pd
from typing import List
import numpy as np
from itertools import chain


class M2C_ReMapId(BaseMid2Cache):
    default_cfg = {
        'share_id_columns': [], # [{c1, c2, c3}]
        'ignore_columns': {"oder_id:token"}
    }

    def __init__(self, m2c_cfg) -> None:
         super().__init__(m2c_cfg)
    
    def _check_params(self):
        for feat in self.m2c_cfg['ignore_columns']: assert 'token' in feat # 确保均是token特征

        # 检查share_id_columns合理性
        t = list(chain(*self.m2c_cfg['share_id_columns']))
        assert len(t) == len(set(t)), 'groups in share_id_columns should be disjoint'

    def process(self, **kwargs):
        super().process(**kwargs)

        feats = self.get_all_columns([v for v in kwargs.values() if type(v) is pd.DataFrame])
        feats = list(filter(lambda x: x.split(":")[-1] in {'token', 'token_seq'}, feats)) # 过滤掉非token特征
        feats = list(filter(lambda x: x not in self.m2c_cfg['ignore_columns'], feats)) # 过滤掉指定token特征
        feats_group = list(set(feats) - set(chain(*self.m2c_cfg['share_id_columns'])))
        feats_group = [set([v]) for v in feats_group]
        feats_group.extend(self.m2c_cfg['share_id_columns'])
        kwargs['feats_group'] = feats_group

        kwargs['lbe_dict'] = {}
        for feats in feats_group:
            col_arr = self.get_specific_column_into_arr(feats, kwargs.values())
            lbe = LabelEncoder().fit(col_arr)

            for v in kwargs.values():
                if type(v) is pd.DataFrame:
                    for col in feats:
                        if col in v.columns:
                            if col.split(":")[-1] == 'token':
                                v[col] = lbe.transform(v[col])
                            elif col.split(":")[-1] == 'token_seq':
                                v[col] = v[col].apply(lambda x: lbe.transform(x).tolist())
                            else:
                                raise ValueError("unsupport type of the feat: {col}")
            for f in feats:
                kwargs['lbe_dict'][f] = lbe
        return kwargs
                        
    @staticmethod
    def get_specific_column_into_arr(columns, df_list: List[pd.DataFrame]):
        col_list = []
        for v in df_list:
            if type(v) is pd.DataFrame:
                for col in columns:
                    if col in v.columns:
                        if col.split(":")[-1] == 'token':
                            col_list.append(v[col].to_numpy())
                        elif col.split(":")[-1] == 'token_seq':
                            col_list.extend(v[col].to_list())
                        else:
                            raise ValueError("unsupport type of the feat: {col}")
        col_arr = np.concatenate(col_list)
        return col_arr
    
    @staticmethod
    def get_all_columns(df_list: List[pd.DataFrame]):
        feats = set()
        for v in df_list: feats |= set(v.columns.tolist())
        return feats



from ..common import BaseMid2Cache
import torch
import numpy as np


class M2C_DIMKT_OP(BaseMid2Cache):
    default_cfg = {}

    def process(self, **kwargs):
        self.df_Q = kwargs['df_exer']
        dt_info = kwargs['dt_info']
        self.num_q = dt_info['exer_count']
        self.num_c = dt_info['cpt_count']

        df_train_folds = kwargs['df_train_folds']

        kwargs['q_diff_list'] = []
        kwargs['c_diff_list'] = []
        for train_dict in df_train_folds:
            self.train_dict = train_dict
            self.compute_difficulty()
            kwargs['q_diff_list'].append(self.q_dif)
            kwargs['c_diff_list'].append(self.c_dif)
        return kwargs

    def compute_difficulty(self):
        q_dict = dict(zip(self.df_Q['exer_id:token'], self.df_Q['cpt_seq:token_seq']))
        qd={}
        qd_count={}
        cd={}
        cd_count={}
        exer_ids = self.train_dict['exer_seq:token_seq']
        label_ids = self.train_dict['label_seq:float_seq']
        mask_ids = self.train_dict['mask_seq:token_seq']
        for ii, ee in enumerate(exer_ids):
            for i, e in enumerate(ee):
                tmp_mask = mask_ids[ii, i]
                if tmp_mask != 0:
                    tmp_exer = exer_ids[ii, i]
                    tmp_label = label_ids[ii, i]
                    cpt = (q_dict[tmp_exer])[0]
                    cd[cpt] = cd.get(cpt, 0) + tmp_label
                    cd_count[cpt] = cd_count.get(cpt, 0) + 1
                    if tmp_exer in qd:
                        qd[tmp_exer] = qd[tmp_exer] + tmp_label
                        qd_count[tmp_exer] = qd_count[tmp_exer]+1
                    else:
                        qd[tmp_exer] =  tmp_label
                        qd_count[tmp_exer] = 1
                else:
                    break


        self.q_dif = np.ones(self.num_q)
        self.c_dif = np.ones(self.num_c)
        for k,v in qd.items():
            self.q_dif[k] = int((qd[k]/qd_count[k])*100)+1
        for k,v in cd.items():
            self.c_dif[k] = int((cd[k]/cd_count[k])*100)+1
        self.q_dif = torch.tensor(self.q_dif).unsqueeze(1)
        self.c_dif = torch.tensor(self.c_dif).unsqueeze(1)



r"""
DIMKT
##########################################

Reference:
    Shuanghong Shen et al. "Assessing Student’s Dynamic Knowledge State by Exploring the Question Difficulty Effect" in SIGIR 2022.

Reference Code:
    https://github.com/pykt-team/pykt-toolkit/blob/main/pykt/models/dimkt.py

"""
from torch.autograd import Variable
from torch.nn import Embedding, Linear, Sigmoid, Tanh, Dropout

from ..gd_basemodel import GDBaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F
import numpy as np
from torch.nn.init import xavier_uniform_, constant_

class DIMKT(GDBaseModel):
    r"""
    DIMKT

    default_cfg:
       'emb_size': 128  # dimension of embedding
       'dropout_rate': 0.2      # dropout rate
       'difficult_levels': 100+2         # difficulty level of the exercises
    """
    default_cfg = {
        'emb_size': 128,
        'difficult_levels': 100+2,
        'dropout':0.2
    }

    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.num_q = self.datatpl_cfg['dt_info']['exer_count']
        self.num_c = self.datatpl_cfg['dt_info']['cpt_count']
        self.emb_size = self.modeltpl_cfg['emb_size']
        self.dropout = self.modeltpl_cfg['dropout']
        self.difficult_levels = self.modeltpl_cfg['difficult_levels']
        
    def build_model(self):
        self.sigmoid = Sigmoid()
        self.tanh = Tanh()
        self.dropout = Dropout(self.dropout)
        self.knowledge = Variable(torch.randn(1, self.emb_size), requires_grad=True)

        self.q_emb = Embedding(self.num_q , self.emb_size, device=self.device, padding_idx=0)
        self.c_emb = Embedding(self.num_c , self.emb_size, device=self.device, padding_idx=0)
        self.sd_emb = Embedding(self.difficult_levels , self.emb_size, device=self.device, padding_idx=0)
        self.qd_emb = Embedding(self.difficult_levels , self.emb_size, device=self.device, padding_idx=0)
        self.a_emb = Embedding(2, self.emb_size, device=self.device)

        self.linear_1 = Linear(4 * self.emb_size, self.emb_size)
        self.linear_2 = Linear(1 * self.emb_size, self.emb_size)
        self.linear_3 = Linear(1 * self.emb_size, self.emb_size)
        self.linear_4 = Linear(2 * self.emb_size, self.emb_size)
        self.linear_5 = Linear(2 * self.emb_size, self.emb_size)
        self.linear_6 = Linear(4 * self.emb_size, self.emb_size)

    def forward(self, exer_seq, label_seq,  **kwargs):
        self.num_steps = exer_seq.shape[1]-1
        self.bs = len(exer_seq)
        q = exer_seq[:,:-1]
        c = kwargs['cpt_unfold_seq'][:,:-1]
        sd = kwargs['cd_seq'][:,:-1].int()
        qd = kwargs['qd_seq'][:,:-1].int()
        a = label_seq[:,:-1].int()
        q_emb = self.q_emb(Variable(q))
        c_emb = self.c_emb(Variable(c))
        sd_emb = self.sd_emb(Variable(sd))
        qd_emb = self.qd_emb(Variable(qd))
        a_emb = self.a_emb(Variable(a))

        qshft = exer_seq[:, 1:]
        cshft = kwargs['cpt_unfold_seq'][:, 1:]
        sdshft = kwargs['cd_seq'][:, 1:].int()
        qdshft = kwargs['qd_seq'][:, 1:].int()
        target_q = self.q_emb(Variable(qshft))
        target_c = self.c_emb(Variable(cshft))
        target_sd = self.sd_emb(Variable(sdshft))
        target_qd = self.qd_emb(Variable(qdshft))

        input_data = torch.cat((q_emb, c_emb, sd_emb, qd_emb), -1)
        input_data = self.linear_1(input_data)

        target_data = torch.cat((target_q, target_c, target_sd, target_qd), -1)
        target_data = self.linear_1(target_data)

        shape = list(sd_emb.shape)
        padd = torch.zeros(shape[0], 1, shape[2], device=self.device)
        sd_emb = torch.cat((padd, sd_emb), 1)
        slice_sd_embedding = sd_emb.split(1, dim=1)

        shape = list(a_emb.shape)
        padd = torch.zeros(shape[0], 1, shape[2], device=self.device)
        a_emb = torch.cat((padd, a_emb), 1)
        slice_a_embedding = a_emb.split(1, dim=1)

        shape = list(input_data.shape)
        padd = torch.zeros(shape[0], 1, shape[2], device=self.device)
        input_data = torch.cat((padd, input_data), 1)
        slice_input_data = input_data.split(1, dim=1)

        qd_emb = torch.cat((padd, qd_emb), 1)
        slice_qd_embedding = qd_emb.split(1, dim=1)

        k = self.knowledge.repeat(self.bs, 1).to(self.device)

        h = list()
        for i in range(1, self.num_steps + 1):
            sd_1 = torch.squeeze(slice_sd_embedding[i], 1)
            a_1 = torch.squeeze(slice_a_embedding[i], 1)
            qd_1 = torch.squeeze(slice_qd_embedding[i], 1)
            input_data_1 = torch.squeeze(slice_input_data[i], 1)

            qq = k - input_data_1

            gates_SDF = self.linear_2(qq)
            gates_SDF = self.sigmoid(gates_SDF)
            SDFt = self.linear_3(qq)
            SDFt = self.tanh(SDFt)
            SDFt = self.dropout(SDFt)

            SDFt = gates_SDF * SDFt

            x = torch.cat((SDFt, a_1), -1)
            gates_PKA = self.linear_4(x)
            gates_PKA = self.sigmoid(gates_PKA)

            PKAt = self.linear_5(x)
            PKAt = self.tanh(PKAt)

            PKAt = gates_PKA * PKAt

            ins = torch.cat((k, a_1, sd_1, qd_1), -1)
            gates_KSU = self.linear_6(ins)
            gates_KSU = self.sigmoid(gates_KSU)

            k = gates_KSU * k + (1 - gates_KSU) * PKAt

            h_i = torch.unsqueeze(k, dim=1)
            h.append(h_i)

        output = torch.cat(h, axis=1)
        logits = torch.sum(target_data * output, dim=-1)
        y = self.sigmoid(logits)

        return y

    @torch.no_grad()
    def predict(self, **kwargs):
        y_pd = self(**kwargs)
        # y_pd = y_pd[:, :-1]
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = None
        if kwargs.get('label_seq', None) is not None:
            y_gt = kwargs['label_seq'][:, 1:]
            y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        return {
            'y_pd': y_pd,
            'y_gt': y_gt
        }

    def get_main_loss(self, **kwargs):
        y_pd= self(**kwargs)
        # y_pd = y_pd[:, :-1]
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = kwargs['label_seq'][:, 1:]
        y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        loss = F.binary_cross_entropy(
            input=y_pd, target=y_gt
        )
        return {
            'loss_main': loss,
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)



from .raw2mid import BaseRaw2Mid
import pandas as pd
import json


class R2M_AAAI_2023(BaseRaw2Mid):
    def get_df(self, df_f):
        records = []
        exerid_dic = {}  # 用来检查重复的习题
        ex_uid = -1
        for index, row in df_f.iterrows():  # 遍历csv所有行
            #         row_indexs.append(index)
            user_id = row["uid"]
            exer_id_s = row["questions"]
            score_s = row["responses"]
            time_s = row["timestamps"]
            #     print(exer_id_s)
            #     print(score_s)
            e_s = exer_id_s.split(",")
            s_s = score_s.split(",")
            t_s = time_s.split(",")
            if ex_uid != user_id:  # 另外一个学生了
                exerid_dic = {}
            for index, e in enumerate(e_s):
                e = int(e)
                if e == -1: break
                re_dict = {}  # 用来保存一条交互记录
                if not e in exerid_dic.keys():
                    exerid_dic[e] = 1
                    re_dict["user_id"] = user_id
                    re_dict["exer_id"] = e
                    re_dict["score"] = int(s_s[index])
                    re_dict["time"] = t_s[index]
                    # for i in n_q:
                    #     if i["exer_id"] == e:
                    #         re_dict["knowledge_code"]= i["knowledge_code"]
                    #         re_dict['content'] = i["text"]
                    #         break
                    # df.loc[len(df)] = [re_dict["user_id"],re_dict["exer_id"],re_dict["score"],re_dict["time"]]
                    records.append([re_dict["user_id"], re_dict["exer_id"], re_dict["score"], re_dict["time"]])
        df = pd.DataFrame(records, columns=['user_id', 'exer_id', 'label', 'start_timestamp'])
        return df

    def get_kw_text(self, q, key2id):
        n_q = []
        for k, v in q.items():  #
            t = {}
            t["exer_id"] = key2id["questions"][k]

            #     "concept_routes": ["2273----1425----637----297", "121----1425----637----297"],
            concept_routes = v["concept_routes"]
            ls_kw = []
            for kw in concept_routes:
                s = kw.split("----")[-1]
                ls_kw.append(key2id["concepts"][s])
            t["knowledge_code"] = list(set(ls_kw))
            t["text"] = v["content"]
            n_q.append(t)
        return n_q

    def process(self):
        df_ = pd.read_csv(f"{self.rawpath}/train_valid_sequences.csv", low_memory=False)
        df_f = df_[df_["fold"] == 0]
        with open(f"{self.rawpath}/questions.json", encoding='utf8') as i_f:
            q = json.load(i_f)
        with open(f"{self.rawpath}/keyid2idx.json", encoding='utf8') as i_f:
            key2id = json.load(i_f)

        n_q = self.get_kw_text(q, key2id)
        df = self.get_df(df_f)
        grouped = df.groupby('user_id', as_index=False)

        # 定义一个函数来修改每个group的DataFrame
        def modify_group(group):
            group.sort_values(['start_timestamp'], inplace=True)
            group['order_id'] = range(len(group))
            return group

        # 使用apply函数来应用修改函数到每个group的DataFrame
        df_modified = grouped.apply(modify_group)
        df = df_modified.reset_index(drop=True)
        df_inter = df[['exer_id', 'user_id', 'start_timestamp', 'order_id', 'label']]
        df_inter = df_inter.rename(
            columns={'user_id': 'stu_id:token', 'exer_id': 'exer_id:token', 'label': 'label:float',
                     'start_timestamp': 'start_timestamp:float',
                     'order_id': 'order_id:float'})
        df_exer = df[['exer_id']]
        df_exer = df_exer.drop_duplicates(subset=['exer_id'])
        records = []
        for e in df_exer['exer_id'].to_list():
            re_dict = {}
            for i in n_q:
                if i["exer_id"] == e:
                    re_dict["knowledge_code"] = i["knowledge_code"]
                    re_dict['content'] = i["text"]
                    break
            records.append([e, re_dict["knowledge_code"], re_dict['content']])
        df_exer = pd.DataFrame(records, columns=['exer_id:token', 'cpt_seq:token_seq', 'content:token_seq'])

        # cpt_ls = []
        df_exer['cpt_seq:token_seq'] = df_exer['cpt_seq:token_seq'].apply(lambda x: ",".join([str(i) for i in x]))
        df_exer['content:token_seq'] = df_exer['content:token_seq'].apply(lambda x: ",".join([str(i) for i in x]))
        # ser_cpt = df['cpt_seq'].astype(str).apply(lambda x: [int(i) for i in x.split(',')])
        # ser_cpt.to_list()
        # cpt_set = set(cpt_ls)
        # # 此处统计数据集，保存到cfg对象中
        # cfg.exer_count = len(df['exer_id'].unique())
        # cfg.stu_count = len(df['user_id'].unique())
        # cfg.cpt_count = len(cpt_set)
        # cfg.interaction_count = len(df_inter)

        df_inter.to_csv(f"{self.midpath}/{self.dt}.inter.csv", index=False, encoding='utf-8')
        df_exer.to_csv(f"{self.midpath}/{self.dt}.exer.csv", index=False, encoding='utf-8')



from ..common import EduDataTPL
import torch
import numpy as np
import pandas as pd


class DKTDSCDataTPL(EduDataTPL):
    default_cfg = {
        'mid2cache_op_seq': ["M2C_KCAsExer", 'M2C_Label2Int', 'M2C_ReMapId', 'M2C_BuildSeqInterFeats','M2C_RandomDataSplit4KT', "M2C_DKTDSC_OP"],
    }

    def __getitem__(self, index):
        dic = super().__getitem__(index)
        # dic['cluster'] = self.cluster[(int(dic['stu_id']), int(dic['seg_seq']))]
        step = len(dic['seg_seq'])
        stu_id = dic['stu_id'].repeat(step)
        # cluster_df = pd.DataFrame([[list(each)] for each in torch.cat((stu_id.unsqueeze(1), dic['seg_seq'].unsqueeze(1)), dim=1).numpy()], columns=['stu_seg_id'])
        # result = pd.merge(cluster_df, self.cluster, on = ['stu_seg_id']).reset_index(drop=True)
        # cluster_id_tensor = torch.Tensor(result['cluster_id'].values)
        # dic['cluster'] = cluster_id_tensor
        dic['cluster'] = np.ones_like(dic['exer_seq'])
        for i in range(step):
            try:
                dic['cluster'][i] = self.cluster.get((int(stu_id[i]), int(dic['seg_seq'][i])))
            except:
                dic['cluster'][i] = 0
        dic['cluster'] = torch.from_numpy(dic['cluster'])
        return dic


    def set_info_for_fold(self, fold_id):
        super().set_info_for_fold(fold_id)
        self.cluster = self.final_kwargs['cluster_list'][fold_id]



from ..common import BaseMid2Cache
import networkx as nx
import numpy as np
from itertools import combinations, chain
import heapq
from collections import defaultdict


class M2C_CDGK_OP(BaseMid2Cache):
    default_cfg = {
        'subgraph_count': 1, 
    }

    def process(self, **kwargs):
        cpt_count = kwargs['dt_info']['cpt_count']
        exer_count = kwargs['dt_info']['exer_count']
        df_exer = kwargs['df_exer']
        subgraph_count = self.m2c_cfg['subgraph_count']

        if subgraph_count == 1:
            kwargs['n_group_of_cpt'] = np.ones(shape=(cpt_count, ), dtype=np.int64)
            kwargs['gid2exers'] = {0: np.arange(exer_count)}
            kwargs['dt_info']['n_cpt_group'] = 1
            return kwargs

        G = self.get_G_from_cpt_concurrence(df_exer, cpt_count, exer_count)
        number_connected_components = nx.number_connected_components(G)

        cpt_group_list = None
        if subgraph_count == number_connected_components or subgraph_count < 0:
            cpt_group_list = nx.connected_components(G)
        elif subgraph_count < number_connected_components:
            # 小根堆，合并，记录合并索引
            subgraphs = list(nx.connected_components(G))
            cpt_group_list = self._merge_nodes_list(subgraphs, subgraph_count)
        else:
            raise NotImplementedError

        cpt2groups = defaultdict(list)
        for gid, conn_comp in enumerate(cpt_group_list):
            for cpt_id in conn_comp: cpt2groups[cpt_id].append(gid)
   
        kwargs['n_group_of_cpt'] = np.array([len(v) for v in cpt2groups.values()], dtype=np.int64)
        kwargs['gid2exers'] = defaultdict(set)
        for _, tmp_df in df_exer[['exer_id:token', 'cpt_seq:token_seq']].iterrows():
            exer_id = tmp_df['exer_id:token']
            cpt_seq = tmp_df['cpt_seq:token_seq']

            for cpt_id in cpt_seq:
                for gid in cpt2groups[cpt_id]:
                    kwargs['gid2exers'][gid].add(exer_id)
        kwargs['gid2exers'] = {gid: np.array(list(exers)) for gid, exers in kwargs['gid2exers'].items()}
        kwargs['dt_info']['n_cpt_group'] = len(kwargs['gid2exers'])

        return kwargs
    

    @staticmethod
    def get_G_from_cpt_concurrence(df_exer, cpt_count, exer_count):
        G = nx.Graph()
        G.add_nodes_from(np.arange(cpt_count))
        df_exer = df_exer.set_index('exer_id:token')
        for exer_id in range(exer_count):
            for edge in combinations(df_exer.loc[exer_id].tolist()[0], 2):
                G.add_edge(edge[0], edge[1])
        return G

    @staticmethod
    def _merge_nodes_list(nodes_list, graph_num):
        """
        合并节点集合，每次选择大小最小的两个集合进行合并

        参数:
        nodes_list (list): 输入的节点集合列表，列表内每一个元素也是一个节点集合
        graph_num (int): 指定返回的列表长度

        返回:
        list: 合并后的节点集合列表
        """
        # 检查输入参数
        if graph_num <= 0 or graph_num > len(nodes_list):
            raise ValueError("graph_num 必须大于0且不超过 nodes_list 的长度")

        # 使用小根堆保存节点集合及其大小
        heap = [(len(nodes), nodes) for nodes in nodes_list]
        heapq.heapify(heap)

        while len(heap) > graph_num:
            # 选择大小最小的两个节点集合进行合并
            len1, nodes1 = heapq.heappop(heap)
            len2, nodes2 = heapq.heappop(heap)
            merged_nodes = nodes1 | nodes2  # 使用集合的并运算合并集合
            heapq.heappush(heap, (len(merged_nodes), merged_nodes))

        # 将合并后的节点集合从小根堆中取出
        merged_nodes_list = [heapq.heappop(heap)[1] for _ in range(len(heap))]

        return merged_nodes_list



from .M2C_EERNN_OP import M2C_EERNN_OP
from .M2C_LPKT_OP import M2C_LPKT_OP
from .M2C_DKTForget_OP import M2C_DKTForget_OP
from .M2C_DKTDSC_OP import M2C_DKTDSC_OP
from .M2C_DIMKT_OP import M2C_DIMKT_OP
from .M2C_QDKT_OP import M2C_QDKT_OP
from .M2C_CL4KT_OP import M2C_CL4KT_OP
from .M2C_MGCD_OP import M2C_MGCD_OP
from .M2C_RCD_OP import M2C_RCD_OP
from .M2C_CDGK_OP import M2C_CDGK_OP



import torch
import torch.nn as nn
import torch.nn.functional as F
from edustudio.model import GDBaseModel
from edustudio.model.utils.components import MLP, PosMLP
import math
import math
from torch.autograd import Function


class MarginLossZeroOne(nn.Module):
    def __init__(self, margin=0.5, reduction: str = 'mean') -> None:
        assert reduction in ['mean', 'sum', 'none']
        super().__init__()
        self.margin = margin
        self.reduction = reduction

    def forward(self, pos_pd, neg_pd):
        logits = self.margin - (pos_pd - neg_pd)
        logits[logits < 0] = 0.0
        if self.reduction == 'mean':
            return logits.mean()
        elif self.reduction == 'sum':
            return logits.sum()
        else:
            return logits


class DCD(GDBaseModel):
    default_cfg = {
        'EncoderUserHidden': [512],
        'EncoderItemHidden': [512],
        'lambda_main': 1.0,
        'lambda_q': 1.0,
        'align_margin_loss_kwargs': {'margin': 0.7, 'topk': 2, "d1":1, 'margin_lambda': 0.5, 'norm': 1, 'norm_lambda': 1.0, 'start_epoch': 1},
        'sampling_type': 'mws',
        'b_sample_type': 'gumbel_softmax',
        'b_sample_kwargs': {'tau': 1.0, 'hard': True},
        'bernoulli_prior_p': 0.1,
        'bernoulli_prior_auto': False,
        'align_type': 'mse_margin',
        'alpha_user': 0.0,
        'alpha_item': 0.0,
        'gamma_user': 1.0,
        'gamma_item': 1.0,
        'beta_user': 0.0,
        'beta_item': 0.0,
        'g_beta_user': 1.0,
        'g_beta_item': 1.0,
        'disc_scale': 10,
        'pred_dnn_units': [256, 128],
        'pred_dropout_rate': 0.5,
        'pred_activation': 'sigmoid',
        'interact_type': 'irt_wo_disc',
    }

    def build_cfg(self):
        self.user_count = self.datatpl_cfg['dt_info']['stu_count']
        self.item_count = self.datatpl_cfg['dt_info']['exer_count']
        self.cpt_count = self.datatpl_cfg['dt_info']['cpt_count']

    def add_extra_data(self, **kwargs):
        self.Q_mat = kwargs['missing_Q_mat'].to(self.device)
        self.dict_cpt_relation = {
            k:torch.LongTensor(v).to(self.device) for k,v in kwargs['dict_cpt_relation'].items()
        }
        self.interact_mat = kwargs['interact_mat'].to(self.device).float()

    def build_model(self):
        self.EncoderUser = MLP(
            input_dim=self.item_count,
            output_dim=self.cpt_count * 2,
            dnn_units=self.modeltpl_cfg['EncoderUserHidden']
        )
        self.EncoderItem = MLP(
            input_dim=self.user_count,
            output_dim=self.cpt_count,
            dnn_units=self.modeltpl_cfg['EncoderItemHidden']
        )

        self.EncoderItemDiff = MLP(
            input_dim=self.user_count,
            output_dim=self.cpt_count * 2,
            dnn_units=self.modeltpl_cfg['EncoderItemHidden']
        )


        self.ItemDisc = nn.Embedding(self.item_count, 1)
        self.pd_net = PosMLP(
            input_dim=self.cpt_count, output_dim=1, activation=self.modeltpl_cfg['pred_activation'],
            dnn_units=self.modeltpl_cfg['pred_dnn_units'], dropout_rate=self.modeltpl_cfg['pred_dropout_rate']
        )

        self.margin_loss_zero_one = MarginLossZeroOne(reduction='none', margin=self.modeltpl_cfg['align_margin_loss_kwargs']['margin'])

        self.user_dist = NormalDistUtil()
        self.item_dist = BernoulliUtil(p=self.modeltpl_cfg['bernoulli_prior_p'], stgradient=True)
        self.item_dist_diff = NormalDistUtil()
    
    def get_align_item_loss(self, item_emb, item_idx):
        if self.modeltpl_cfg['align_type'] == 'mse_margin':
            flag = self.Q_mat[item_idx, :].sum(dim=1) > 0
            left_emb = item_emb[~flag]
            p = self.modeltpl_cfg['align_margin_loss_kwargs']['norm']
            t_loss = torch.norm(left_emb, dim=0, p=p).pow(p).sum()
            if left_emb.shape[0] != 0 and self.callback_list.curr_epoch >= self.modeltpl_cfg['align_margin_loss_kwargs']['start_epoch']:
                # topk_idx = torch.topk(left_emb, self.modeltpl_cfg['align_margin_loss_kwargs']['topk']).indices
                # bottomk_idx = torch.ones_like(left_emb).scatter(1, topk_idx, 0).nonzero()[:, 1].reshape(-1, left_emb.size(1) - topk_idx.size(1))
                # pos = torch.gather(left_emb, 1, topk_idx[:,[-1]])
                # neg = torch.gather(left_emb, 1, bottomk_idx[:,torch.randperm(bottomk_idx.shape[1],dtype=torch.long)[0:int(bottomk_idx.shape[1]*0.5)]])
                topk_idx = torch.topk(left_emb, self.modeltpl_cfg['align_margin_loss_kwargs']['topk']+1).indices
                pos = torch.gather(left_emb, 1, topk_idx[:,0:self.modeltpl_cfg['align_margin_loss_kwargs']['d1']])
                neg = torch.gather(left_emb, 1, topk_idx[:,[-1]])
                margin_loss = self.margin_loss_zero_one(pos, neg).mean(dim=1).sum()
            else:
                margin_loss = torch.tensor(0.0).to(self.device)
            return {
                "mse_loss": F.mse_loss(item_emb[flag], self.Q_mat[item_idx[flag], :].float(), reduction='sum'),
                "margin_loss": margin_loss,
                "norm_loss": t_loss,
            }
        elif self.modeltpl_cfg['align_type'] == 'mse_margin_mean':
            flag = self.Q_mat[item_idx, :].sum(dim=1) > 0
            left_emb = item_emb[~flag]
            p = self.modeltpl_cfg['align_margin_loss_kwargs']['norm']
            t_loss = torch.norm(left_emb, dim=0, p=p).pow(p).sum()
            if left_emb.shape[0] != 0 and self.callback_list.curr_epoch >= self.modeltpl_cfg['align_margin_loss_kwargs']['start_epoch']:
                # topk_idx = torch.topk(left_emb, self.modeltpl_cfg['align_margin_loss_kwargs']['topk']).indices
                # bottomk_idx = torch.ones_like(left_emb).scatter(1, topk_idx, 0).nonzero()[:, 1].reshape(-1, left_emb.size(1) - topk_idx.size(1))
                # pos = torch.gather(left_emb, 1, topk_idx[:,[-1]])
                # neg = torch.gather(left_emb, 1, bottomk_idx[:,torch.randperm(bottomk_idx.shape[1],dtype=torch.long)[0:int(bottomk_idx.shape[1]*0.5)]])
                topk_idx = torch.topk(left_emb, self.modeltpl_cfg['align_margin_loss_kwargs']['topk']+1).indices
                bottomk_idx = torch.topk(-left_emb, left_emb.shape[1] -  self.modeltpl_cfg['align_margin_loss_kwargs']['topk']).indices
                pos = torch.gather(left_emb, 1, topk_idx[:,0:self.modeltpl_cfg['align_margin_loss_kwargs']['d1']])
                neg = torch.gather(left_emb, 1, bottomk_idx).mean(dim=1)
                margin_loss = self.margin_loss_zero_one(pos, neg).mean(dim=1).sum()
            else:
                margin_loss = torch.tensor(0.0).to(self.device)
            return {
                "mse_loss": F.mse_loss(item_emb[flag], self.Q_mat[item_idx[flag], :].float(), reduction='sum'),
                "margin_loss": margin_loss,
                "norm_loss": t_loss,
            }
        else:
            raise ValueError(f"Unknown align type: {self.modeltpl_cfg['align_type']}")


    def decode(self, user_emb, item_emb, item_emb_diff, item_id, **kwargs):
        if self.modeltpl_cfg['interact_type'] == 'irt_wo_disc':
            return ((user_emb - item_emb_diff)*item_emb).sum(dim=1)
        elif self.modeltpl_cfg['interact_type'] == 'irt':
            item_disc = self.ItemDisc(item_id).sigmoid() #* self.modeltpl_cfg['disc_scale']
            return ((user_emb - item_emb_diff)*item_emb*item_disc).sum(dim=1)
        elif self.modeltpl_cfg['interact_type'] == 'ncdm':
            item_disc = self.ItemDisc(item_id).sigmoid()# * self.modeltpl_cfg['disc_scale']
            input = (user_emb - item_emb_diff)*item_emb*item_disc
            return self.pd_net(input).flatten()
        elif self.modeltpl_cfg['interact_type'] == 'mf':
            return ((user_emb.sigmoid()*item_emb)*(item_emb*item_emb_diff)).sum(dim=1)
        elif self.modeltpl_cfg['interact_type'] == 'mirt': # 就是mf加了个disc
            item_disc = self.ItemDisc(item_id).sigmoid() #* self.modeltpl_cfg['disc_scale']
            return ((user_emb.sigmoid()*item_emb)*(item_emb*item_emb_diff)).sum(dim=1) + item_disc.flatten()
        else:
            raise NotImplementedError

    def forward(self, users, items, labels):
        user_unique, user_unique_idx = users.unique(sorted=True, return_inverse=True)
        item_unique, item_unique_idx = items.unique(sorted=True, return_inverse=True)

        user_mix = self.EncoderUser(self.interact_mat[user_unique, :])
        user_mu, user_logvar = torch.chunk(user_mix, 2, dim=-1)
        user_emb_ = self.user_dist.sample(user_mu, user_logvar)
        user_emb = user_emb_[user_unique_idx, :]

        item_mu = self.EncoderItem(self.interact_mat[:, item_unique].T).sigmoid()
        item_emb_ = self.item_dist.sample(None, item_mu, type_=self.modeltpl_cfg['b_sample_type'], **self.modeltpl_cfg['b_sample_kwargs'])
        item_emb = item_emb_[item_unique_idx, :]


        item_diff_mix = self.EncoderItemDiff(self.interact_mat[:, item_unique].T)
        item_mu_diff, item_logvar_diff = torch.chunk(item_diff_mix, 2, dim=-1)
        item_emb_diff_ = self.item_dist_diff.sample(item_mu_diff, item_logvar_diff)
        item_emb_diff = item_emb_diff_[item_unique_idx, :]

        loss_main = F.binary_cross_entropy_with_logits(self.decode(user_emb, item_emb, item_emb_diff, item_id=items), labels, reduction='sum') # 重构 loss
        align_loss_dict = self.get_align_item_loss(item_mu, item_unique)
        # align_loss_dict_diff = self.get_align_item_loss(item_mu_diff, item_unique)

        user_terms = self.get_tcvae_terms(user_emb_, params=(user_mu, user_logvar), dist=self.user_dist, dataset_size=self.user_count)
        item_terms = self.get_tcvae_terms(item_emb_, params=item_mu, dist=self.item_dist, dataset_size=self.item_count)
        item_terms_diff = self.get_tcvae_terms(item_emb_diff_, params=(item_mu_diff, item_logvar_diff), dist=self.item_dist_diff, dataset_size=self.item_count)
        
        return {
            'loss_main': loss_main * self.modeltpl_cfg['lambda_main'],
            'loss_mse': align_loss_dict['mse_loss'] * self.modeltpl_cfg['lambda_q'],
            'loss_margin': align_loss_dict['margin_loss'] * self.modeltpl_cfg['align_margin_loss_kwargs']['margin_lambda'],
            'loss_norm': align_loss_dict['norm_loss'] * self.modeltpl_cfg['align_margin_loss_kwargs']['norm_lambda'],
            # 'loss_mse_diff': align_loss_dict_diff['mse_loss'] * self.modeltpl_cfg['lambda_q'],
            # 'loss_margin_diff': align_loss_dict_diff['margin_loss'] * self.modeltpl_cfg['align_margin_loss_kwargs']['margin_lambda'],
            # 'loss_norm_diff': align_loss_dict_diff['norm_loss'] * self.modeltpl_cfg['align_margin_loss_kwargs']['norm_lambda'],
            
            'user_MI': user_terms['MI'] * self.modeltpl_cfg['alpha_user'],
            'user_TC': user_terms['TC'] * self.modeltpl_cfg['beta_user'],
            'user_TC_G': user_terms['TC_G'] * self.modeltpl_cfg['g_beta_user'],
            'user_KL': user_terms['KL'] * self.modeltpl_cfg['gamma_user'],
            'item_MI': item_terms['MI'] * self.modeltpl_cfg['alpha_item'],
            'item_TC': item_terms['TC'] * self.modeltpl_cfg['beta_item'],
            'item_TC_G': item_terms['TC_G'] * self.modeltpl_cfg['g_beta_item'],
            'item_KL': item_terms['KL'] * self.modeltpl_cfg['gamma_item'],
            
            'item_MI_diff': item_terms_diff['MI'] * self.modeltpl_cfg['alpha_item'],
            'item_TC_diff': item_terms_diff['TC'] * self.modeltpl_cfg['beta_item'],
            'item_TC_G_diff': item_terms_diff['TC_G'] * self.modeltpl_cfg['g_beta_item'],
            'item_KL_diff': item_terms_diff['KL'] * self.modeltpl_cfg['gamma_item'],
        }

    def get_tcvae_terms(self, z, params, dist, dataset_size):
        batch_size, latent_dim = z.shape

        if isinstance(dist, NormalDistUtil):
            mu, logvar = params
            zero = torch.FloatTensor([0.0]).to(self.device)
            logpz = dist.log_density(X=z, MU=zero, LOGVAR=zero).sum(dim=1)
            logqz_condx = dist.log_density(X=z, MU=mu, LOGVAR=logvar).sum(dim=1)
            _logqz = dist.log_density(
                z.reshape(batch_size, 1, latent_dim),
                mu.reshape(1, batch_size, latent_dim),
                logvar.reshape(1, batch_size, latent_dim)
            ) # _logqz的第(i,j,k)个元素, P(z(n_i)_k|n_j)
        elif isinstance(dist, BernoulliUtil):
            logpz = dist.log_density(z, params=None).sum(dim=1)
            logqz_condx = dist.log_density(z, params=params).sum(dim=1)
            # _logqz = torch.stack([dist.log_density(z, params[i,:]) for i in range(batch_size)],dim=1)
            _logqz = dist.log_density(z.reshape(batch_size, 1, latent_dim), params=params.reshape(1, batch_size, latent_dim), is_check=False)
        else:
            raise ValueError("unknown base class of dist")

        if self.modeltpl_cfg['sampling_type'] == 'mws':
            # minibatch weighted sampling
            logqz_prodmarginals = (torch.logsumexp(_logqz, dim=1, keepdim=False) - math.log(batch_size * dataset_size)).sum(1)
            logqz = (torch.logsumexp(_logqz.sum(dim=2), dim=1, keepdim=False) - math.log(batch_size * dataset_size))
            logqz_group_list = []
            if hasattr(self, 'dict_cpt_relation'):
                for gid, group_idx in self.dict_cpt_relation.items():
                    logqz_group_list.append(
                        (torch.logsumexp(_logqz[:,:,group_idx].sum(dim=2), dim=1, keepdim=False) - math.log(batch_size * dataset_size))
                    )
                logqz_group = torch.vstack(logqz_group_list).T.sum(dim=1)
        elif self.modeltpl_cfg['sampling_type'] == 'mss':
            logiw_mat = self._log_importance_weight_matrix(z.shape[0], dataset_size).to(z.device)
            logqz = torch.logsumexp(
                logiw_mat + _logqz.sum(dim=-1), dim=-1
            )  # MMS [B]
            logqz_prodmarginals = (
                torch.logsumexp(
                    logiw_mat.reshape(z.shape[0], z.shape[0], -1) + _logqz,
                    dim=1,
                )
            ).sum(
                dim=-1
            )
            logqz_group_list = []
            if hasattr(self, 'dict_cpt_relation'):
                for gid, group_idx in self.dict_cpt_relation.items():
                    logqz_group_list.append(
                       (
                        torch.logsumexp(
                            logiw_mat.reshape(z.shape[0], z.shape[0], -1) + _logqz[:,:,group_idx], dim=1,
                        )).sum(dim=-1)
                    )
                logqz_group = torch.vstack(logqz_group_list).T.sum(dim=1)

        else:
            raise ValueError("Unknown Sampling Type")
        
        IndexCodeMI = logqz_condx - logqz
        TC = logqz - logqz_prodmarginals
        TC_G = (logqz - logqz_group).mean() if hasattr(self, 'dict_cpt_relation') else torch.FloatTensor([0.0]).to(self.device)
        DW_KL = logqz_prodmarginals - logpz
        return {
            'MI': IndexCodeMI.mean(),
            'TC': TC.mean(),
            'TC_G': TC_G,
            'KL': DW_KL.mean()
        }


    @torch.no_grad()
    def predict(self, stu_id, exer_id, **kwargs):
        users = stu_id
        items = exer_id
        user_emb = None
        if users is None:
            user_mix = self.EncoderUser(self.interact_mat)
            user_emb, _ = torch.chunk(user_mix, 2, dim=-1)
        else:
            user_mix = self.EncoderUser(self.interact_mat[users, :])
            user_emb, _ = torch.chunk(user_mix, 2, dim=-1)

        item_emb = None
        if items is None:
            item_emb = self.EncoderItem(self.interact_mat.T).sigmoid()
        else:
            item_emb = self.EncoderItem(self.interact_mat[:, items].T).sigmoid()

        item_emb_diff = None
        if items is None:
            item_emb_diff_mix = self.EncoderItemDiff(self.interact_mat.T)
            item_emb_diff, _ = torch.chunk(item_emb_diff_mix, 2, dim=-1)
        else:
            item_emb_diff_mix = self.EncoderItemDiff(self.interact_mat[:, items].T)
            item_emb_diff, _ = torch.chunk(item_emb_diff_mix, 2, dim=-1)

        return {"y_pd":self.decode(user_emb, item_emb, item_emb_diff, item_id=items).sigmoid()}

    @torch.no_grad()
    def get_stu_status(self, users=None):
        user_emb = None
        if users is None:
            user_mix = self.EncoderUser(self.interact_mat)
            user_emb, _ = torch.chunk(user_mix, 2, dim=-1)
        else:
            user_mix = self.EncoderUser(self.interact_mat[users, :])
            user_emb, _ = torch.chunk(user_mix, 2, dim=-1)
        
        return user_emb

    @torch.no_grad()
    def get_exer_emb(self, items=None):
        item_emb = None
        if items is None:
            item_emb = self.EncoderItem(self.interact_mat.T)
        else:
            item_emb = self.EncoderItem(self.interact_mat[:, items].T)
        
        return item_emb.sigmoid()

    def _log_importance_weight_matrix(self, batch_size, dataset_size):
        """Compute importance weigth matrix for MSS
        Code from (https://github.com/rtqichen/beta-tcvae/blob/master/vae_quant.py)
        """

        N = dataset_size
        M = batch_size - 1
        strat_weight = (N - M) / (N * M)
        W = torch.Tensor(batch_size, batch_size).fill_(1 / M)
        W.view(-1)[:: M + 1] = 1 / N
        W.view(-1)[1 :: M + 1] = strat_weight
        W[M - 1, 0] = strat_weight
        return W.log()

    def get_main_loss(self, **kwargs):
        stu_id = kwargs['stu_id']
        exer_id = kwargs['exer_id']
        label = kwargs['label']
        return self(stu_id, exer_id, label)

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)



eps=1e-8


class STHeaviside(Function):
    @staticmethod
    def forward(ctx, x):
        y = torch.zeros(x.size()).type_as(x)
        y[x >= 0] = 1
        return y

    @staticmethod
    def backward(ctx, grad_output):
        return grad_output


class NormalDistUtil(object):
    @staticmethod
    def log_density(X, MU, LOGVAR):
        """ compute log pdf of normal distribution

        Args:
            X (_type_): sample point
            MU (_type_): mu of normal dist
            LOGVAR (_type_): logvar of normal dist
        """
        norm = - 0.5 * (math.log(2 * math.pi) + LOGVAR)
        log_density = norm - 0.5 * ((X - MU).pow(2) * torch.exp(-LOGVAR))
        return log_density

    @staticmethod
    def kld(MU:float, LOGVAR:float, mu_move):
        """compute KL divergence between X and Normal Dist whose (mu, var) equals to (mu_move, 1)

        Args:
            MU (float): _description_
            VAR (float): _description_
            mu_move (_type_): _description_
        """

        return 0.5 * (LOGVAR.exp() - LOGVAR + MU.pow(2) - 2 * mu_move * MU + mu_move ** 2 - 1)

    @staticmethod
    def sample(mu, logvar):
        std = torch.exp(logvar/2)
        eps = torch.randn_like(std)
        return mu + std * eps


class BernoulliUtil(nn.Module):
    """Samples from a Bernoulli distribution where the probability is given
    by the sigmoid of the given parameter.
    """

    def __init__(self, p=0.5, stgradient=False):
        super().__init__()
        p = torch.Tensor([p])
        self.p = torch.log(p / (1 - p) + eps)
        self.stgradient = stgradient

    def _check_inputs(self, size, ps):
        if size is None and ps is None:
            raise ValueError(
                'Either one of size or params should be provided.')
        elif size is not None and ps is not None:
            if ps.ndimension() > len(size):
                return ps.squeeze(-1).expand(size)
            else:
                return ps.expand(size)
        elif size is not None:
            return self.p.expand(size)
        elif ps is not None:
            return ps
        else:
            raise ValueError(
                'Given invalid inputs: size={}, ps={})'.format(size, ps))

    def _sample_logistic(self, size):
        u = torch.rand(size)
        l = torch.log(u + eps) - torch.log(1 - u + eps)
        return l

    def default_sample(self, size=None, params=None):
        presigm_ps = self._check_inputs(size, params)
        logp = F.logsigmoid(presigm_ps)
        logq = F.logsigmoid(-presigm_ps)
        l = self._sample_logistic(logp.size()).type_as(presigm_ps)
        z = logp - logq + l
        b = STHeaviside.apply(z)
        return b if self.stgradient else b.detach()

    def sample(self, size=None, params=None,type_='gumbel_softmax', **kwargs):
        if type_ == 'default':
            return self.default_sample(size, params)
        elif type_ == 'gumbel_softmax':
            tau = kwargs.get('tau', 1.0)
            hard = kwargs.get('hard', True)
            ext_params = torch.log(torch.stack([1 - params, params],dim=2) + eps)
            return F.gumbel_softmax(logits=ext_params, tau=tau, hard=hard)[:,:,-1]
        else:
            raise ValueError(f"Unknown Type of sample: {type_}")

    def log_density(self, sample, params=None, is_check=True):
        if is_check:
            presigm_ps = self._check_inputs(sample.size(), params).type_as(sample)
        else:
            presigm_ps = params
        p = (torch.sigmoid(presigm_ps) + eps) * (1 - 2 * eps)
        logp = sample * torch.log(p + eps) + (1 - sample) * torch.log(1 - p + eps)
        return logp

    def get_params(self):
        return self.p

    @property
    def nparams(self):
        return 1

    @property
    def ndim(self):
        return 1

    @property
    def is_reparameterizable(self):
        return self.stgradient

    def __repr__(self):
        tmpstr = self.__class__.__name__ + ' ({:.3f})'.format(
            torch.sigmoid(self.p.data)[0])
        return tmpstr



from edustudio.model import GDBaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F


class MF(GDBaseModel):
    default_cfg = {
        'emb_dim': 32,
        'reg_user': 0.0,
        'reg_item': 0.0
    }

    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        self.emb_size = self.modeltpl_cfg['emb_dim']
        self.reg_user = self.modeltpl_cfg['reg_user']
        self.reg_item = self.modeltpl_cfg['reg_item']

    def build_model(self):
        self.user_emb = nn.Embedding(
            num_embeddings=self.n_user,
            embedding_dim=self.emb_size
        )
        self.item_emb = nn.Embedding(
            num_embeddings=self.n_item,
            embedding_dim=self.emb_size
        )

    def forward(self, user_idx: torch.LongTensor, item_idx: torch.LongTensor):
        assert len(user_idx.shape) == 1 and len(item_idx.shape) == 1 and user_idx.shape[0] == item_idx.shape[0]
        return torch.einsum("ij,ij->i", self.user_emb(user_idx), self.item_emb(item_idx)).sigmoid()

    @torch.no_grad()
    def predict(self, stu_id, exer_id, **kwargs):
        return {
            'y_pd': self(stu_id, exer_id).flatten(),
        }

    def get_main_loss(self, **kwargs):
        stu_id = kwargs['stu_id']
        exer_id = kwargs['exer_id']
        label = kwargs['label']
        pd = self(stu_id, exer_id).flatten()
        loss = F.binary_cross_entropy(input=pd, target=label)
        return {
            'loss_main': loss
        }
        
    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)


import torch
import torch.nn as nn
import numpy as np
import torch.nn.functional as F
from torch.autograd import Variable

# Graph-based Knowledge Tracing: Modeling Student Proficiency Using Graph Neural Network.
# For more information, please refer to https://dl.acm.org/doi/10.1145/3350546.3352513
# Author: jhljx
# Email: jhljx8918@gmail.com

def sample_gumbel(shape, eps=1e-10):
    """
    NOTE: Stolen from https://github.com/ethanfetaya/NRI/blob/master/utils.py
    Sample from Gumbel(0, 1)
    based on
    https://github.com/ericjang/gumbel-softmax/blob/3c8584924603869e90ca74ac20a6a03d99a91ef9/Categorical%20VAE.ipynb ,
    """
    U = torch.rand(shape).float()
    return - torch.log(eps - torch.log(U + eps))


def gumbel_softmax_sample(logits, tau=1, eps=1e-10, dim=-1):
    """
    NOTE: Stolen from https://github.com/ethanfetaya/NRI/blob/master/utils.py
    Draw a sample from the Gumbel-Softmax distribution
    based on
    https://github.com/ericjang/gumbel-softmax/blob/3c8584924603869e90ca74ac20a6a03d99a91ef9/Categorical%20VAE.ipynb
    """
    gumbel_noise = sample_gumbel(logits.size(), eps=eps)
    if logits.is_cuda:
        gumbel_noise = gumbel_noise.cuda()
    y = logits + Variable(gumbel_noise)
    return F.softmax(y / tau, dim=dim)


def gumbel_softmax(logits, tau=1, hard=False, eps=1e-10, dim=-1):
    """
    NOTE: Stolen from https://github.com/ethanfetaya/NRI/blob/master/utils.py
    Sample from the Gumbel-Softmax distribution and optionally discretize.
    Args:
      logits: [batch_size, n_class] unnormalized log-probs
      tau: non-negative scalar temperature
      hard: if True, take argmax, but differentiate w.r.t. soft sample y
    Returns:
      [batch_size, n_class] sample from the Gumbel-Softmax distribution.
      If hard=True, then the returned sample will be one-hot, otherwise it will
      be a probability distribution that sums to 1 across classes
    Constraints:
    - this implementation only works on batch_size x num_features tensor for now
    based on
    https://github.com/ericjang/gumbel-softmax/blob/3c8584924603869e90ca74ac20a6a03d99a91ef9/Categorical%20VAE.ipynb
    """
    y_soft = gumbel_softmax_sample(logits, tau=tau, eps=eps, dim=dim)
    if hard:
        shape = logits.size()
        _, k = y_soft.data.max(-1)
        # this bit is based on
        # https://discuss.pytorch.org/t/stop-gradients-for-st-gumbel-softmax/530/5
        y_hard = torch.zeros(*shape)
        if y_soft.is_cuda:
            y_hard = y_hard.cuda()
        y_hard = y_hard.zero_().scatter_(-1, k.view(shape[:-1] + (1,)), 1.0)
        # this cool bit of code achieves two things:
        # - makes the output value exactly one-hot (since we add then
        #   subtract y_soft value)
        # - makes the gradient equal to y_soft gradient (since we strip
        #   all other gradients)
        y = Variable(y_hard - y_soft.data) + y_soft
    else:
        y = y_soft
    return y


def kl_categorical(preds, log_prior, concept_num, eps=1e-16):
    kl_div = preds * (torch.log(preds + eps) - log_prior)
    return kl_div.sum() / (concept_num * preds.size(0))


def kl_categorical_uniform(preds, concept_num, num_edge_types, add_const=False, eps=1e-16):
    kl_div = preds * torch.log(preds + eps)
    if add_const:
        const = np.log(num_edge_types)
        kl_div += const
    return kl_div.sum() / (concept_num * preds.size(0))


def nll_gaussian(preds, target, variance, add_const=False):
    # pred: [concept_num, embedding_dim]
    # target: [concept_num, embedding_dim]
    neg_log_p = ((preds - target) ** 2 / (2 * variance))
    if add_const:
        const = 0.5 * np.log(2 * np.pi * variance)
        neg_log_p += const
    return neg_log_p.mean()


class VAELoss(nn.Module):

    def __init__(self, concept_num, edge_type_num=2, prior=False, var=5e-5):
        super(VAELoss, self).__init__()
        self.concept_num = concept_num
        self.edge_type_num = edge_type_num
        self.prior = prior
        self.var = var

    def forward(self, ec_list, rec_list, z_prob_list, log_prior=None):
        time_stamp_num = len(ec_list)
        loss = 0
        for time_idx in range(time_stamp_num):
            output = rec_list[time_idx]
            target = ec_list[time_idx]
            prob = z_prob_list[time_idx]
            loss_nll = nll_gaussian(output, target, self.var)
            if self.prior:
                assert log_prior is not None
                loss_kl = kl_categorical(prob, log_prior, self.concept_num)
            else:
                loss_kl = kl_categorical_uniform(prob, self.concept_num, self.edge_type_num)
            if time_idx == 0:
                loss = loss_nll + loss_kl
            else:
                loss = loss + loss_nll + loss_kl
        return loss / time_stamp_num



from ..common import BaseMid2Cache
import pandas as pd
import numpy as np
from sklearn.preprocessing import LabelEncoder


class M2C_LPKT_OP(BaseMid2Cache):
    default_cfg = {
        'answer_time_unit': 'ms', # ['ms', 's', 'minute']
        'start_timestamp_unit': 'ms', # ['ms', 's', 'minute']
    }

    def process(self, **kwargs):
        df_train_folds = kwargs['df_train_folds']
        df_valid_folds = kwargs['df_valid_folds']
        df_test_folds = kwargs['df_test_folds']

        self.dt_info = kwargs['dt_info']
        self.dt_info['answer_time_count_list'] = []
        self.dt_info['interval_time_count_list'] = []

        for idx, (train_dict, test_dict) in enumerate(zip(df_train_folds, df_test_folds)):
            self.train_dict = train_dict
            self.test_dict = test_dict
            if df_valid_folds is not None and len(df_valid_folds) > 0:
                self.val_dict = df_valid_folds[idx]
                
            self.construct_answer_time()
            self.construct_interval_time()

        return kwargs

    def construct_answer_time(self):
        # 1. train, val, test中所有answer_time变成秒
        curr_unit = self.m2c_cfg['answer_time_unit']
        answer_time = []
        answer_time.append(self.answer_time2seconds(self.train_dict['answer_time_seq:float_seq'], curr_unit))
        self.train_dict['answer_time_seq:float_seq'] = answer_time[-1]
        if self.val_dict is not None:
            answer_time.append(self.answer_time2seconds(self.val_dict['answer_time_seq:float_seq'], curr_unit))
            self.val_dict['answer_time_seq:float_seq'] = answer_time[-1]
        answer_time.append(self.answer_time2seconds(self.test_dict['answer_time_seq:float_seq'], curr_unit))
        self.test_dict['answer_time_seq:float_seq'] = answer_time[-1]

        # 2. 离散化
        at_lbe = LabelEncoder()
        at_lbe.fit(np.concatenate(answer_time).flatten())
        self.train_dict['answer_time_seq:float_seq'] = at_lbe.transform(
            self.train_dict['answer_time_seq:float_seq'].flatten()).reshape(self.train_dict['answer_time_seq:float_seq'].shape)
        if self.val_dict is not None:
            self.val_dict['answer_time_seq:float_seq'] = at_lbe.transform(
                self.val_dict['answer_time_seq:float_seq'].flatten()).reshape(self.val_dict['answer_time_seq:float_seq'].shape)
        self.test_dict['answer_time_seq:float_seq'] = at_lbe.transform(
            self.test_dict['answer_time_seq:float_seq'].flatten()).reshape(self.test_dict['answer_time_seq:float_seq'].shape)

        self.dt_info['answer_time_count_list'].append(len(at_lbe.classes_))
        

    def construct_interval_time(self):
        # 1. 计算train, val, test中的interval_time, 将interval time变成分钟
        curr_unit = self.m2c_cfg['start_timestamp_unit']
        interval_time = []
        train_it = self.train_dict['start_timestamp_seq:float_seq'] - np.concatenate([self.train_dict['start_timestamp_seq:float_seq'][:,[0]], self.train_dict['start_timestamp_seq:float_seq'][:,0:-1]], axis=1)
        train_it[train_it < 0] = 0
        train_it = self.interval_time2minutes(train_it, curr_unit)
        train_it[train_it > 43200] = 43200
        interval_time.append(train_it)
        test_it = self.test_dict['start_timestamp_seq:float_seq'] - np.concatenate([self.test_dict['start_timestamp_seq:float_seq'][:,[0]], self.test_dict['start_timestamp_seq:float_seq'][:,0:-1]], axis=1)
        test_it[test_it < 0] = 0
        test_it = self.interval_time2minutes(test_it, curr_unit)
        test_it[test_it > 43200] = 43200
        interval_time.append(test_it)
        if self.val_dict is not None:
            val_it = self.val_dict['start_timestamp_seq:float_seq'] - np.concatenate([self.val_dict['start_timestamp_seq:float_seq'][:,[0]], self.val_dict['start_timestamp_seq:float_seq'][:,0:-1]], axis=1)
            val_it[val_it < 0] = 0
            val_it = self.interval_time2minutes(val_it, curr_unit)
            val_it[val_it > 43200] = 43200
            interval_time.append(val_it)
        
        # 2. 离散化
        it_lbe = LabelEncoder()
        it_lbe.fit(np.concatenate(interval_time).flatten())
        self.train_dict['interval_time_seq'] = it_lbe.transform(train_it.flatten()).reshape(train_it.shape)
        if self.val_dict is not None:
            self.val_dict['interval_time_seq'] = it_lbe.transform(val_it.flatten()).reshape(val_it.shape)
        self.test_dict['interval_time_seq'] = it_lbe.transform(test_it.flatten()).reshape(test_it.shape)

        self.dt_info['interval_time_count_list'].append(len(it_lbe.classes_))

    @staticmethod
    def answer_time2seconds(tensor, curr_unit):
        if curr_unit == 'ms':
            return tensor // 1000
        elif curr_unit == 's':
            return tensor
        elif curr_unit == 'minute':
            return tensor * 60
        else:
            ValueError(f"unsupported answer_time_unit: {curr_unit}")
    
    @staticmethod
    def interval_time2minutes(tensor, curr_unit):
        if curr_unit == 'ms':
            return tensor // (1000 * 60)
        elif curr_unit == 's':
            return tensor // 60
        elif curr_unit == 'minute':
            return tensor
        else:
            ValueError(f"unsupported answer_time_unit: {curr_unit}")



# -*- coding: utf-8 -*-
# @Author : Xiangzhi Chen
# @Github : kervias

from .callbacks.callback import Callback
from .callbacks.history import History
from .callbacks.baseLogger import BaseLogger
from typing import Sequence
import torch.nn as nn
from functools import reduce
from .modeState import ModeState
import logging


class CallbackList(object):
    def __init__(self, callbacks: Sequence[Callback], model: nn.Module, 
                 add_logger: bool = True, logger: logging.Logger = None):
        """

        :param callbacks:
        :param model:
        :param add_history:
        :param add_logger:
        """
        self.callbacks = list(filter(lambda x: isinstance(x, Callback), callbacks))
        self._add_default_callbacks(add_logger)
        self._reg_logger(logger or logging.getLogger())
        self._reg_model(model)
        self._reg_callback_list()
        self._set_mode_state(ModeState.START)

        self.curr_epoch = None

    def _reg_model(self, model: nn.Module):
        self.model = model
        for cb in self.callbacks:
            cb.set_model(model)

    def _add_default_callbacks(self, add_logger: bool):
        self._logger = reduce(
            lambda x, y: x or y, map(lambda x: x if isinstance(x, BaseLogger) else None, self.callbacks)
        )
        if add_logger and not self._logger:
            self._logger = BaseLogger()
            self.callbacks.append(self._logger)

    def _set_mode_state(self, mode_state: ModeState):
        self.mode_state = mode_state
        for cb in self.callbacks:
            cb.set_state(mode_state=mode_state)

    def _reg_logger(self, logger: logging.Logger):
        self.logger = logger
        for cb in self.callbacks:
            cb.set_logger(logger)

    def _reg_callback_list(self):
        for cb in self.callbacks:
            cb.set_callback_list(self)

    def on_train_begin(self, logs=None, **kwargs):
        assert self.mode_state == ModeState.START
        self.curr_epoch = 1
        self._set_mode_state(ModeState.TRAINING)
        for cb in self.callbacks:
            cb.on_train_begin(logs, **kwargs)

    def on_train_end(self, logs=None, **kwargs):
        assert self.mode_state == ModeState.TRAINING
        self.curr_epoch = None
        self._set_mode_state(ModeState.END)
        for cb in self.callbacks:
            cb.on_train_end(logs, **kwargs)

    def on_epoch_begin(self, epoch, logs=None, **kwargs):
        assert self.mode_state == ModeState.TRAINING
        assert epoch == self.curr_epoch
        self.model.train()
        for cb in self.callbacks:
            cb.on_epoch_begin(epoch, logs, **kwargs)

    def on_epoch_end(self, epoch, logs=None, **kwargs):
        assert self.mode_state == ModeState.TRAINING
        assert epoch == self.curr_epoch
        for cb in self.callbacks:
            cb.on_epoch_end(epoch, logs, **kwargs)
        self.curr_epoch += 1

    def on_train_batch_begin(self, batch, logs=None, **kwargs):
        for cb in self.callbacks:
            cb.on_train_batch_begin(batch, logs, **kwargs)

    def on_train_batch_end(self, batch, logs=None, **kwargs):
        for cb in self.callbacks:
            cb.on_train_batch_end(batch, logs, **kwargs)
